# This file is part of HYLOA - HYsteresis LOop Analyzer.
# Copyright (C) 2024 Francesco Zeno Costanzo

# HYLOA is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# HYLOA is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with HYLOA. If not, see <https://www.gnu.org/licenses/>.


"""
Code for creation of the log panel in the main window
"""
from PyQt5.QtCore import QTimer
from PyQt5.QtGui import QTextCursor
from PyQt5.QtWidgets import QPlainTextEdit


class LogWindow(QPlainTextEdit):
    ''' Class to handle the log panel
    '''
    def __init__(self, app_instance):
        super().__init__()
        self.app_instance = app_instance
        self.setReadOnly(True)
        self.setStyleSheet("background-color: white; color: black; font-family: monospace;")
        self.timer = QTimer(self)
        self.timer.timeout.connect(self.update_log)
        self.timer.start(1000)    # Update each seconds

        self.last_line_count = 0  # track last line shown

    def update_log(self):
        ''' Function that update the window
        '''
        if not self.app_instance.logger_path:
            return

        try:
            try :
                with open(self.app_instance.logger_path, "r", encoding="utf-8") as f:
                    lines = f.readlines()
            except UnicodeDecodeError:
                with open(self.app_instance.logger_path, "r", encoding="cp1252") as f:
                    lines = f.readlines()
            
            if len(lines) == self.last_line_count:
                return

            # Show only new lines
            new_lines = lines[self.last_line_count:]
            self.last_line_count = len(lines)

            cleaned = [line.split(" INFO - ")[-1] for line in new_lines]

            # Scroll on bottom
            scrollbar = self.verticalScrollBar()
            at_bottom = scrollbar.value() == scrollbar.maximum()

            self.appendPlainText("".join(cleaned))

            if at_bottom:
                self.moveCursor(QTextCursor.End)

        except Exception as e:
            self.appendPlainText(f"\n[Error] {e}")



# This file is part of HYLOA - HYsteresis LOop Analyzer.
# Copyright (C) 2024 Francesco Zeno Costanzo

# HYLOA is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# HYLOA is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with HYLOA. If not, see <https://www.gnu.org/licenses/>.


"""
Code for logger setup
"""
import logging
from PyQt5.QtWidgets import QFileDialog, QMessageBox


def setup_logging(log_file):
    '''
    Configures logging to the specified file.

    Parameters
    ----------
    log_file : str
        Path to the log file.
    '''
    try:
        # Remove any existing handlers
        for handler in logging.root.handlers[:]:
            logging.root.removeHandler(handler)

        logging.basicConfig(
            filename=log_file,
            level=logging.INFO,
            encoding="utf-8",
            format="%(asctime)s -  %(name)s - %(levelname)s - %(message)s",
            datefmt="%Y-%m-%d %H:%M:%S"
        )
        logging.info("Start of log session.")
        logging.info(f"Logging configured: write on {log_file}")
    except Exception as e:
        raise Exception(f"Error while configuring logging: {e}")


def start_logging(app_instance, parent_widget=None):
    '''
    Starts the logging session by selecting a file and configuring the logger.

    Parameters
    ----------
    app_instance : MainApp
        Instance of the main application class
    parent_widget : QWidget
        The parent window for dialog placement (optional)
    '''
    if app_instance.logger is not None:
        QMessageBox.information(
            parent_widget,
            "Info",
            "Logger already configured."
        )
        return
    # Inform the user that the file will be appended to if it already exists
    QMessageBox.information(
        parent_widget,
        "Info",
        "If you choose an existing file, writing will be queued, without overwriting."
    )

    # Let the user choose the file
    log_file, _ = QFileDialog.getSaveFileName(
        parent_widget,
        "Select the log file",
        "",
        "Log Files (*.log);;Tutti i file (*)"
    )

    if log_file:
        try:
            setup_logging(log_file)
            app_instance.logger = logging.getLogger(__name__)
            app_instance.logger.info("Logging configured successfully.")
            app_instance.logger_path = log_file

            QMessageBox.information(
                parent_widget,
                "Logging Started",
                f"The log will be written on the file:\n{log_file}"
            )

            app_instance.open_default_panels()
            
        except Exception as e:
            QMessageBox.critical(
                parent_widget,
                "Error",
                f"Error while configuring logging:\n{e}"
            )
    else:
        QMessageBox.critical(
            parent_widget,
            "Error",
            "Please select a valid file for the log."
        )


# This file is part of HYLOA - HYsteresis LOop Analyzer.
# Copyright (C) 2024 Francesco Zeno Costanzo

# HYLOA is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# HYLOA is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with HYLOA. If not, see <https://www.gnu.org/licenses/>.


"""
Code to save a session (i.e. alla dta loaded all plot crated and so on) in .pkl files
"""
import os
import sys
import gzip
import pickle
import logging
from datetime import datetime


from PyQt5.QtCore import QTimer

from PyQt5.QtWidgets import (
    QMdiSubWindow, QMessageBox, QFileDialog
)

from hyloa.utils.logging_setup import setup_logging

from hyloa.gui.plot_window import PlotSubWindow
from hyloa.gui.plot_window import PlotControlWidget
from hyloa.gui.worksheet import WorksheetWindow

def save_current_session(app_instance, parent_widget=None):
    '''
    Save the current session to a HYLOA session file (.hyloa)
    Session files are intended to be opened only by HYLOA.
    Backward compatibility is guaranteed across minor versions.

    Parameters
    ----------
    app_instance : MainApp
        Main application instance containing the session data.
    parent_widget : QWidget or None
        Optional parent for the dialog windows.
    '''
    if app_instance.logger is None:
        QMessageBox.critical(parent_widget, "Error", "Cannot start analysis without starting log.")
        return

    file_path, _ = QFileDialog.getSaveFileName(
        parent_widget,
        "Save session",
        "",
        "HYLOA Session (*.hyloa)"
    )

    if not file_path:
        QMessageBox.warning(parent_widget, "Warning", "No file selected for saving.")
        return
    
    if not file_path.lower().endswith(".hyloa"):
        file_path += ".hyloa"


    try:
        # Build the dictionary for saving data 
        session_data = {

            "__hyloa__":           True,
            "file_format_version": 1,
            "app_version":         getattr(app_instance, "version", "unknown"),
            "created_on":          datetime.now().isoformat(),
            "platform":            sys.platform,

            "dataframes":   app_instance.dataframes,
            "header_lines": app_instance.header_lines,
            "logger_path":  app_instance.logger_path,
            "log_filename": os.path.basename(app_instance.logger_path),
            "fit_results":  app_instance.fit_results,
            "number_plots": app_instance.number_plots,

            "plot_widgets": {
                idx: {
                    "selected_pairs": [
                        (
                            f_combo.currentText(),
                            x_combo.currentText(),
                            y_combo.currentText()
                        )
                        for f_combo, x_combo, y_combo in widget.selected_pairs
                    ],
                    "plot_customizations": widget.plot_customizations.copy()
                }
                for idx, widget in app_instance.plot_widgets.items()
            },
            "plot_names": {
                idx : name for idx, name in app_instance.plot_names.items()
            },
            "control_windows_geometry": {
                idx: {
                    "x": sub.x(),
                    "y": sub.y(),
                    "width":  sub.width(),
                    "height": sub.height(),
                    "minimized": sub.isMinimized()
                }
                for idx, sub in app_instance.plot_subwindows.items()
            },
            "plot_windows_geometry": {
                idx: {
                    "x": fig_sub.x(),
                    "y": fig_sub.y(),
                    "width":  fig_sub.width(),
                    "height": fig_sub.height(),
                    "minimized": fig_sub.isMinimized()
                }
                for idx, fig_sub in app_instance.figure_subwindows.items()
            },
            "worksheets": {
                idx: {
                    "name":    app_instance.worksheet_names.get(idx, f"Worksheet {idx}"),
                    "content": app_instance.worksheet_windows[idx].to_session_data()
                }
                for idx in app_instance.worksheet_windows
            },

        }

        with gzip.open(file_path, "wb") as f:
            pickle.dump(session_data, f, protocol=pickle.HIGHEST_PROTOCOL)

        QMessageBox.information(parent_widget, "Session saved",
                                f"Session saved in file:\n{file_path}")
    except Exception as e:
        QMessageBox.critical(parent_widget, "Error", f"Error while saving:\n{e}")


def load_previous_session(app_instance, parent_widget=None):
    '''
    Load a previously saved session from a .hyloa file.

    Parameters
    ----------
    app_instance : MainApp
        Main application instance to load the session into.
    parent_widget : QWidget or None
        Optional parent for the dialog windows.
    '''
    options = QFileDialog.Options()
    file_path, _ = QFileDialog.getOpenFileName(
        parent_widget,
        "Load session",
        "",
        "HYLOA Session (*.hyloa);;Legacy Pickle (*.pkl)",
        options=options
    )

    if not file_path:
        QMessageBox.warning(parent_widget, "Error", "No file selected for session loading.")
        return

    try:
        try:
            with gzip.open(file_path, "rb") as f:
                session_data = pickle.load(f)
        except OSError:
            # --- fallback to legacy pickle ---
            with open(file_path, "rb") as f:
                session_data = pickle.load(f)

        # Reload attributes of main app instance
        app_instance.dataframes     = session_data.get("dataframes", [])
        app_instance.header_lines   = session_data.get("header_lines", [])
        app_instance.fit_results    = session_data.get("fit_results", {})
        app_instance.number_plots   = session_data.get("number_plots", 0)

        try :
            app_instance.logger_path = session_data.get("logger_path", None)
            # Recreate the logger
            setup_logging(app_instance.logger_path)
        except :

            log_filename = session_data.get("log_filename")
            # To ensure compatibility with different OS
            default_dir              = os.path.dirname(file_path)
            reconstructed_path       = os.path.join(default_dir, log_filename)
            app_instance.logger_path = reconstructed_path
            # Recreate the logger
            setup_logging(app_instance.logger_path)

        
        app_instance.logger = logging.getLogger(__name__)

        is_new_format = session_data.get("__hyloa__", False)
        file_version  = session_data.get("file_format_version", 0)

        if is_new_format:
            app_instance.logger.info(
                f"Loaded HYLOA session (format v{file_version})"
            )
        else:
            app_instance.logger.info(
                "Loaded legacy HYLOA session (no signature)"
            )

        app_instance.logger.info("Logger restored from session file.")

        app_instance.open_default_panels()

        # Recreate all plot's control panels
        plot_widgets_data = session_data.get("plot_widgets", {})
        plot_names_data   = session_data.get("plot_names",   {}) 
        
        # Loop only over widget because widgets and names share the same keys
        for idx_str, plot_info in plot_widgets_data.items():
            
            idx       = int(idx_str)
            plot_name = plot_names_data.get(idx, f"Graph {idx}")

            widget = PlotControlWidget(app_instance, idx, plot_name)
            
            for i in reversed(range(widget.pair_layout.count())):
                widget.pair_layout.itemAt(i).widget().setParent(None)
            widget.selected_pairs.clear()

            app_instance.plot_widgets[idx] = widget
            app_instance.plot_names[idx]   = plot_name

            # Retrieve selected pairs ...
            for file_str, x_str, y_str in plot_info.get("selected_pairs", []):
                widget.add_pair(file_text=file_str, x_col=x_str, y_col=y_str)

            # ... and customization
            widget.plot_customizations = plot_info.get("plot_customizations", {})

            # Create sub window for panel
            sub = PlotSubWindow(app_instance, widget, idx)
            app_instance.mdi_area.addSubWindow(sub)
            sub.show()

            # Restore size and position for control panels...
            ctrl_geom = session_data.get("control_windows_geometry", {}).get(idx)
            if ctrl_geom:
                sub.setGeometry(ctrl_geom["x"], ctrl_geom["y"], ctrl_geom["width"], ctrl_geom["height"])
                if ctrl_geom.get("minimized"):
                    sub.showMinimized()
            
            app_instance.plot_subwindows[idx] = sub

            widget.plot() # Force the plot
            # ... and for plot windows
            fig_geom = session_data.get("plot_windows_geometry", {}).get(idx)
            fig_sub  = app_instance.figure_subwindows.get(idx)
            if fig_geom and fig_sub:
                fig_sub.setGeometry(fig_geom["x"], fig_geom["y"], fig_geom["width"], fig_geom["height"])
                if fig_geom.get("minimized"):
                    fig_sub.showMinimized()

        # --- restore worksheets  ---
        worksheet_data = session_data.get("worksheets", {})
        for idx_key, ws_info in worksheet_data.items():
            try:
                idx = int(idx_key)
            except:
                idx = idx_key

            ws_name = ws_info.get("name", f"Worksheet {idx}")

            # Create worksheet instance with name
            ws = WorksheetWindow(app_instance.mdi_area, parent=app_instance.mdi_area, name=ws_name,
                                logger=app_instance.logger, app_instance=app_instance)

            # Add to mdi area BEFORE restoring content (important!)
            app_instance.mdi_area.addSubWindow(ws)

            # Restore content & plots; from_session_data will set geometry and schedule show
            ws.from_session_data(ws_info.get("content", {}))

            # Save references in app_instance
            app_instance.worksheet_windows[idx]    = ws
            app_instance.worksheet_names[idx]      = ws_name
            app_instance.worksheet_subwindows[idx] = ws

            if isinstance(idx, int) and idx > app_instance.number_worksheets:
                app_instance.number_worksheets = idx + 1

            # Ensure it's visible (from_session_data schedules showNormal/showMinimized)
            QTimer.singleShot(0, lambda w=ws: w.show())


        # Rename main window to reflect the loaded session
        base_name = os.path.basename(file_path)
        name, _   = os.path.splitext(base_name)
        app_instance.setWindowTitle(f"HYLOA - {name}")
            
        QMessageBox.information(parent_widget, "Session Loaded",
                                f"Session loaded from file:\n{file_path}")

    except Exception as e:
        QMessageBox.critical(parent_widget, "Error",
                             f"Error while loading session:\n{e}")


# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

import re
import os
import sys
sys.path.insert(0, os.path.abspath(".."))  # Allow import hyloa/

project = 'hyloa'
copyright = '2025, Francesco Zeno Costanzo'
author = 'Francesco Zeno Costanzo'
release = "1.7.0"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",   # Automatic generation of documentation from docstrings 
    "sphinx.ext.napoleon",  # Support for NumPy and Google style docstrings
    "sphinx.ext.viewcode",  # Link to souce code
    "sphinx.ext.autosummary",
]

templates_path = ['_templates']
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']
autosummary_generate = True


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_rtd_theme"
html_static_path = ['_static']



# This file is part of HYLOA - HYsteresis LOop Analyzer.
# Copyright (C) 2024 Francesco Zeno Costanzo

# HYLOA is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# HYLOA is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with HYLOA. If not, see <https://www.gnu.org/licenses/>.


"""
Code to manage the window where you can
run code to make changes to the data
"""

import io
import sys
import numpy as np
from PyQt5.QtWidgets import (
    QWidget, QVBoxLayout, QLabel, QTextEdit
)
from PyQt5.QtCore import Qt

from scipy.special import *
from scipy.optimize import *
import matplotlib.pyplot as plt

class ShellEditor(QTextEdit):
    ''' Class for wrinting in the shell
    '''

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setFontFamily("Courier")
        self.setFontPointSize(10)
        self.setLineWrapMode(QTextEdit.NoWrap)
        self.setText(">>> ")

    def keyPressEvent(self, event):
        ''' Function to handle events from key pressing
        '''
        cursor = self.textCursor() 
        cursor_pos = cursor.position()
        prompt_pos = self.toPlainText().rfind(">>> ") + 4

        # Protection against Backspace and Delete on the prompt
        if cursor_pos <= prompt_pos and event.key() in (Qt.Key_Backspace, Qt.Key_Delete):
            return 

        # Block left arrow cursor movement before prompt
        if cursor_pos <= prompt_pos and event.key() == Qt.Key_Left:
            return

        # If I type before the prompt, I go back to the end of the text
        if cursor_pos < prompt_pos:
            cursor.setPosition(len(self.toPlainText()))
            self.setTextCursor(cursor)

        super().keyPressEvent(event)

    def mousePressEvent(self, event):
        ''' Function to handle events from mouse pressing
        '''
        cursor = self.cursorForPosition(event.pos())
        cursor_pos = cursor.position()
        prompt_pos = self.toPlainText().rfind(">>> ") + 4

        if cursor_pos < prompt_pos:
            # Force cursor after prompt
            cursor.setPosition(len(self.toPlainText()))
            self.setTextCursor(cursor)
        else:
            super().mousePressEvent(event)


class CommandWindow(QWidget):
    ''' 
    Class for all the window, 
    for comand history navigation and evaluation
    '''

    def __init__(self, app_instance):
        super().__init__()
        self.app_instance = app_instance
        self.dataframes = app_instance.dataframes
        self.fit_results = app_instance.fit_results
        self.logger = app_instance.logger

        self.setWindowTitle("Python shell")

        layout = QVBoxLayout(self)
        layout.addWidget(QLabel("Only in-line command", self))

        self.shell_text = ShellEditor(self)
        layout.addWidget(self.shell_text)

        self.local_vars = self._initialize_local_vars()
        self.command_history = []
        self.history_index = -1

        self.shell_text.installEventFilter(self)
        self.shell_text.moveCursor(self.shell_text.textCursor().End)

    def _initialize_local_vars(self):
        ''' Function to initialize the varibales konwn by the shell
        '''
        local_vars = {}

        for idx, df in enumerate(self.dataframes):
            for column in df.columns:
                local_vars[column] = df[column].astype(float).values

        local_vars.update(self.fit_results)
        return local_vars

    def refresh_variables(self):
        ''' Function to update the varibales konwn by the shell
        '''
        self.local_vars.clear()

        for idx, df in enumerate(self.app_instance.dataframes):
            for column in df.columns:
                self.local_vars[column] = df[column].astype(float).values

        self.local_vars.update(self.app_instance.fit_results)


    def eventFilter(self, obj, event):
        ''' Handle switch for navigation in command's history or execute command
        '''
        if obj is self.shell_text:
            if event.type() == event.KeyPress:
                key = event.key()

                if key == Qt.Key_Return:
                    self.execute_command()
                    return True

                elif key == Qt.Key_Up:
                    self.navigate_history(-1)
                    return True

                elif key == Qt.Key_Down:
                    self.navigate_history(1)
                    return True

        return super().eventFilter(obj, event)

    def execute_command(self):
        ''' Function that execute the command
        '''
        text = self.shell_text.toPlainText()
        last_prompt = text.rfind(">>> ")
        command = text[last_prompt + 4:].strip()

        if not command:
            self.shell_text.append(">>> ")
            return

        self.command_history.append(command)
        self.history_index = len(self.command_history)

        if self.logger:
            self.logger.info(f"Execution of the command: {command}")

        old_stdout, old_stderr = sys.stdout, sys.stderr
        sys.stdout = output_capture = io.StringIO()
        sys.stderr = output_capture

        try:
            # First try to execute as expression
            try:
                code = compile(command, "<input>", "eval")
                res  = eval(code, globals(), self.local_vars)

                if res is not None:
                    print(repr(res))
                    self.local_vars["_"] = res
            except SyntaxError:
                # If not possible, execute as statement
                exec(command, globals(), self.local_vars)

            output = output_capture.getvalue()

        except Exception as e:
            output = f"Error: {str(e)}\n"
        finally:
            sys.stdout, sys.stderr = old_stdout, old_stderr

        # Update dataframes with modified variables
        for idx, df in enumerate(self.dataframes):
            for column in df.columns:
                if column in self.local_vars:
                    modified_array = self.local_vars[column]
                    if not np.array_equal(df[column].values, modified_array):
                        df[column] = modified_array

        self.shell_text.append(output)
        self.shell_text.append(">>> ")
        self.shell_text.moveCursor(self.shell_text.textCursor().End)

    def navigate_history(self, direction):
        ''' Function for navigating in all command's history
        '''
        if not self.command_history:
            return

        self.history_index += direction
        self.history_index = max(0, min(self.history_index, len(self.command_history) - 1))

        command = self.command_history[self.history_index]

        text = self.shell_text.toPlainText()
        last_prompt = text.rfind(">>> ")
        new_text = text[:last_prompt + 4] + command
        self.shell_text.setText(new_text)
        self.shell_text.moveCursor(self.shell_text.textCursor().End)
    
    def append_text(self, text):
        cursor = self.shell_text.textCursor()
        cursor.movePosition(cursor.End)
        cursor.insertText(text)
        self.shell_text.setTextCursor(cursor)
        self.shell_text.ensureCursorVisible()




# This file is part of HYLOA - HYsteresis LOop Analyzer.
# Copyright (C) 2024 Francesco Zeno Costanzo

# HYLOA is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# HYLOA is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with HYLOA. If not, see <https://www.gnu.org/licenses/>.

"""
Code for the worksheet's windows.
"""

from PyQt5.QtWidgets import (
    QLineEdit, QWidget, QVBoxLayout, QPushButton, 
    QFrame, QDialog, QLabel, QComboBox, QGridLayout,
    QDialogButtonBox, QStackedWidget, QScrollArea,
    QHBoxLayout, QGroupBox, QFormLayout
)


class ColumnSelectionDialog(QDialog):
    '''
    Dialog for selecting columns to plot.
    '''
    def __init__(self, columns, parent=None):
        '''
        Initialize the column selection dialog.

        Parameters
        ----------
        columns : list of str
            List of column names available for selection.
        parent : QWidget, optional
            The parent widget (default is None).
        '''
        super().__init__(parent)
        self.setWindowTitle("Select columns for Plotting")
        self.setMinimumWidth(500)

        self.columns = columns
        self.curve_rows = []

        main_layout = QVBoxLayout(self)

        # Create a scroll area for curve selections
        self.scroll_area = QScrollArea()
        self.scroll_area.setWidgetResizable(True)

        self.scroll_widget = QWidget()
        self.curves_layout = QVBoxLayout(self.scroll_widget)
        self.curves_layout.setSpacing(10)

        self.scroll_area.setWidget(self.scroll_widget)
        main_layout.addWidget(self.scroll_area)

        # Button to add more curves
        btn_add = QPushButton("Add curve")
        btn_add.clicked.connect(self.add_curve_row)
        main_layout.addWidget(btn_add)

        # OK/Cancel buttons
        buttons = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel)
        buttons.accepted.connect(self.accept)
        buttons.rejected.connect(self.reject)
        main_layout.addWidget(buttons)

        # Add the first curve selection row
        self.add_curve_row()
    
    def add_curve_row(self):
        ''' Add a new row for selecting X, Y, Xerr, Yerr columns.
        '''
        curve_index = len(self.curve_rows) + 1

         # === Main container for one curve ===
        container = QFrame()
        container.setFrameShape(QFrame.StyledPanel)

        container_layout = QVBoxLayout(container)

        # ---- Header row (Curve X + Remove button) ----
        header_layout = QHBoxLayout()

        title_label = QLabel(f"Curve {curve_index}")
        title_label.setStyleSheet("font-weight: bold;")

        remove_btn = QPushButton("Remove")
        remove_btn.setMaximumWidth(80)

        header_layout.addWidget(title_label)
        header_layout.addStretch()
        header_layout.addWidget(remove_btn)

        container_layout.addLayout(header_layout)

        grid  = QGridLayout()

        x_combo = QComboBox(); x_combo.addItems(self.columns)
        y_combo = QComboBox(); y_combo.addItems(self.columns)

        xerr_combo = QComboBox(); xerr_combo.addItem("None"); xerr_combo.addItems(self.columns)
        yerr_combo = QComboBox(); yerr_combo.addItem("None"); yerr_combo.addItems(self.columns)

        # Layout grid (clean and aligned)
        grid.addWidget(QLabel("x:"),     0, 0)
        grid.addWidget(x_combo,          0, 1)

        grid.addWidget(QLabel("y:"),     1, 0)
        grid.addWidget(y_combo,          1, 1)

        grid.addWidget(QLabel("xerr:"),  0, 2)
        grid.addWidget(xerr_combo,       0, 3)

        grid.addWidget(QLabel("yerr:"),  1, 2)
        grid.addWidget(yerr_combo,       1, 3)

        container_layout.addLayout(grid)
        self.curves_layout.addWidget(container)

        remove_btn.clicked.connect(
            lambda: self.remove_curve(container)
        )

        self.curve_rows.append({
            "container" : container,
            "title"     : title_label,
            "x"         : x_combo,
            "y"         : y_combo,
            "xerr"      : xerr_combo,
            "yerr"      : yerr_combo
        })

    def remove_curve(self, container):
        '''Remove a curve container.'''

        # Remove from layout
        container.setParent(None)
        container.deleteLater()

        # Remove from internal list
        self.curve_rows = [
            row for row in self.curve_rows
            if row["container"] != container
        ]

        # Renumber remaining curves
        for i, row in enumerate(self.curve_rows):
            row["title"].setText(f"Curve {i+1}")

    def get_selection(self):
        ''' 
        Get the list of selected columns for all curves.

        Return
        ------
        selections : list
            A list of dictionaries, each containing:
            - x: selected X column
            - y: selected Y column
            - x_err: selected X error column or None 
        '''
        selections = []
        
        for row in self.curve_rows:

            x = row["x"].currentText()
            y = row["y"].currentText()

            x_err = row["xerr"].currentText()
            y_err = row["yerr"].currentText()

            selections.append({
                "x": x,
                "y": y,
                "x_err": None if x_err == "None" else x_err,
                "y_err": None if y_err == "None" else y_err,
            })
        return selections

#===========================================================================================#
#===========================================================================================#
#===========================================================================================#

class ColumnMathDialog(QDialog):
    ''' Dialog for performing arithmetic operations between columns.
    '''
    def __init__(self, columns, parent=None):
        '''
        Initialize the column math dialog.

        Parameters
        ----------
        columns : list of str
            List of column names available for selection.
        parent : QWidget, optional
            The parent widget (default is None).
        '''
        super().__init__(parent)
        self.setWindowTitle("Column Math")

        main_layout = QVBoxLayout(self)

        # Mode selection
        mode_group  = QGroupBox("Operation mode")
        mode_layout = QVBoxLayout()
        self.mode   = QComboBox()
        self.mode.addItems([
            "Arithmetic between columns",
            "Custom expression between columns",
            "Generate linspace",
            "Generate logspace",
            "Generate function from linspace/logspace"
        ])
        mode_layout.addWidget(self.mode)
        mode_group.setLayout(mode_layout)
        main_layout.addWidget(mode_group)


        # Create the stcack area for dynamic widgets
        self.stack = QStackedWidget()
        self.pages = {
            "Arithmetic between columns": self.create_arithmetic_page(columns),
            "Custom expression between columns": self.create_expression_page(columns),
            "Generate linspace": self.create_space_page(),
            "Generate logspace": self.create_space_page(),
            "Generate function from linspace/logspace": self.create_function_page()
        }

        # Add pages to stack
        for page in self.pages.values():
            self.stack.addWidget(page)
        
        content_group  = QGroupBox("Parameters")
        content_layout = QVBoxLayout()
        content_layout.addWidget(self.stack)
        content_group.setLayout(content_layout)

        main_layout.addWidget(content_group)
        
        # Name new column
        name_layout   = QFormLayout()
        self.new_name = QLineEdit()
        self.new_name.setPlaceholderText("New column name")
        name_layout.addRow("New column name:", self.new_name)
        main_layout.addLayout(name_layout)

        # Buttons
        btns = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel)
        btns.accepted.connect(self.accept)
        btns.rejected.connect(self.reject)
        main_layout.addWidget(btns)

        # Connections
        self.mode.currentTextChanged.connect(self.switch_page)

    def switch_page(self, mode):
        '''
        Switch the displayed page based on selected mode.

        Parameters
        ----------
        mode : str
            The selected mode.
        '''
        self.stack.setCurrentWidget(self.pages[mode])

    def create_arithmetic_page(self, columns):
        '''
        Create the page for arithmetic operations between columns.

        Parameters
        ----------
        columns : list of str
            List of column names available for selection.
        
        Returns
        -------
        QWidget
            The created page widget.
        '''
        page = QWidget()
        form = QFormLayout(page)

        self.col_a         = QComboBox(); self.col_a.addItems(columns)
        self.op            = QComboBox(); self.op.addItems(["+", "-", "*", "/", "mean"])
        self.col_b         = QComboBox(); self.col_b.addItems(["<Constant>"] + list(columns))
        self.constant_edit = QLineEdit(); self.constant_edit.setPlaceholderText("Constant value")
        self.col_b.currentTextChanged.connect(self.toggle_constant)

        form.addRow("Column A:", self.col_a)
        form.addRow("Operation:", self.op)
        form.addRow("Column B:", self.col_b)
        form.addRow("Constant:", self.constant_edit)
        return page
    
    def create_expression_page(self, columns):

        page = QWidget()
        form = QFormLayout(page)

        info = QLabel(
            "Write a numpy-compatible expression.\n"
            "Available variables: " + ", ".join(columns)
        )
        info.setWordWrap(True)

        self.expr_edit = QLineEdit()
        self.expr_edit.setPlaceholderText(
            "(col1 - col2) / col3"
        )

        form.addRow(info)
        form.addRow("Expression:", self.expr_edit)

        return page
    
    def create_space_page(self):
        '''
        Create the page for generating linspace or logspace.
        
        Returns
        -------
        QWidget
            The created page widget.
        '''
        page = QWidget()
        form = QFormLayout(page)

        self.start_edit = QLineEdit(); self.start_edit.setPlaceholderText("Start value")
        self.stop_edit  = QLineEdit(); self.stop_edit.setPlaceholderText("Stop value")
        self.num_edit   = QLineEdit(); self.num_edit.setPlaceholderText("Number of points")

        form.addRow("Start:", self.start_edit)
        form.addRow("Stop:", self.stop_edit)
        form.addRow("Number of points:", self.num_edit)
        
        return page
    
    def create_function_page(self):
        '''
        Create the page for generating a function from linspace/logspace.

        Returns
        -------
        QWidget
            The created page widget.
        '''
        page = QWidget()
        form = QFormLayout(page)

        self.space_type = QComboBox()
        self.space_type.addItems(["linspace", "logspace"])

        self.start_edit = QLineEdit(); self.start_edit.setPlaceholderText("Start value")
        self.stop_edit  = QLineEdit(); self.stop_edit.setPlaceholderText("Stop value")
        self.num_edit   = QLineEdit(); self.num_edit.setPlaceholderText("Number of points")
        self.func_edit  = QLineEdit(); self.func_edit.setPlaceholderText("Must be numpy function, e.g. np.sin(x)")

        for label, widget in [
            ("Space type:", self.space_type),
            ("Start:", self.start_edit),
            ("Stop:", self.stop_edit),
            ("Number of points:", self.num_edit),
            ("Function f(x):", self.func_edit)]:

            form.addRow(label, widget)

        return page
        
    def toggle_constant(self, text):
        '''
        Enable/disable constant input based on selection.

        Parameters
        ----------
        text : str
            The current text of the col_b combo box.
        '''
        self.constant_edit.setEnabled(text == "<Constant>")

    def get_selection(self):
        '''
        Get the selected columns, operation, constant, and new column name.
        
        Returns
        -------
        dict
            A dictionary with keys:
            - mode: selected mode
            - col_a: first column
            - op: operation
            - col_b: second column or None
            - const: constant value or None
            - space_type: linspace or logspace
            - start: start value for linspace/logspace
            - stop: stop value for linspace/logspace
            - num: number of points for linspace/logspace
            - func: function for generating new column
            - new_name: name of the new column
        '''
        mode = self.mode.currentText()
        return {
            "mode":       mode,
            "col_a":      getattr(self, "col_a", None).currentText() if hasattr(self, "col_a") else None,
            "op":         getattr(self, "op", None).currentText() if hasattr(self, "op") else None,
            "col_b":      None if not hasattr(self, "col_b") or self.col_b.currentText() == "<Constant>" else self.col_b.currentText(),
            "const":      self.constant_edit.text() if hasattr(self, "constant_edit") else "",
            "expr":       self.expr_edit.text().strip() if hasattr(self, "expr_edit") else "",
            "space_type": self.space_type.currentText() if hasattr(self, "space_type") else "linspace",
            "start":      self.start_edit.text() if hasattr(self, "start_edit") else "",
            "stop":       self.stop_edit.text() if hasattr(self, "stop_edit") else "",
            "num":        self.num_edit.text() if hasattr(self, "num_edit") else "",
            "func":       self.func_edit.text().strip() if hasattr(self, "func_edit") else "",
            "new_name":   self.new_name.text().strip()
        }


# This file is part of HYLOA - HYsteresis LOop Analyzer.
# Copyright (C) 2024 Francesco Zeno Costanzo

# HYLOA is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# HYLOA is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with HYLOA. If not, see <https://www.gnu.org/licenses/>.

"""
test for load and save session
"""

import pickle
import pytest
from unittest.mock import patch, MagicMock

from hyloa.data.session import save_current_session
from hyloa.data.session import load_previous_session

@pytest.fixture
def fake_app_instance():
    """ Fake app_instance with all required attributes.
    """
    app              = MagicMock()
    app.logger       = "dummy_logger"
    app.dataframes   = {"df1": "data"}
    app.header_lines = ["#header"]
    app.logger_path  = "/path/to/log.txt"
    app.fit_results  = {"fit": "results"}
    app.number_plots = 1

    # plot_widgets with selected_pairs and plot_customizations
    combo_mock = MagicMock()
    combo_mock.currentText.side_effect = ["f", "x", "y"]
    plot_widget = MagicMock()
    plot_widget.selected_pairs = [(combo_mock, combo_mock, combo_mock)]
    plot_widget.plot_customizations.copy.return_value = {"style": "custom"}
    app.plot_widgets = {0: plot_widget}

    app.plot_names = {0: "Plot 0"}

    subwindow_mock                          = MagicMock()
    subwindow_mock.x.return_value           = 0
    subwindow_mock.y.return_value           = 0
    subwindow_mock.width.return_value       = 400
    subwindow_mock.height.return_value      = 300
    subwindow_mock.isMinimized.return_value = False

    app.plot_subwindows   = {0: subwindow_mock}
    app.figure_subwindows = {0: subwindow_mock}

    return app


def test_save_fails_without_logger(fake_app_instance):
    fake_app_instance.logger = None

    with patch("hyloa.data.session.QMessageBox.critical") as critical_mock:
        save_current_session(fake_app_instance)
        critical_mock.assert_called_once()
        assert "log" in critical_mock.call_args[0][2]  # check error text includes "log"


def test_save_cancelled_by_user(fake_app_instance):
    with patch("hyloa.data.session.QFileDialog.getSaveFileName", return_value=("", "")), \
         patch("hyloa.data.session.QMessageBox.warning") as warning_mock:
        save_current_session(fake_app_instance)
        warning_mock.assert_called_once()
        assert "No file selected" in warning_mock.call_args[0][2]


def test_save_successful(tmp_path, fake_app_instance):
    test_file = tmp_path / "test_session.hyloa"

    with patch("hyloa.data.session.QFileDialog.getSaveFileName",
               return_value=(str(test_file), "")), \
         patch("hyloa.data.session.pickle.dump") as dump_mock, \
         patch("hyloa.data.session.QMessageBox.information") as info_mock:

        save_current_session(fake_app_instance)

        assert test_file.exists()
        dump_mock.assert_called_once()
        info_mock.assert_called_once()


def test_save_raises_exception(fake_app_instance):
    # Force pickle to raise exception by mocking open
    with patch("hyloa.data.session.QFileDialog.getSaveFileName", return_value=("fake.pkl", "")), \
         patch("hyloa.data.session.QMessageBox.information"), \
         patch("hyloa.data.session.QMessageBox.critical") as critical_mock, \
         patch("builtins.open", side_effect=OSError("fail")):

        save_current_session(fake_app_instance)
        critical_mock.assert_called_once()
        assert "Error while saving" in critical_mock.call_args[0][2]



@patch("hyloa.data.session.QMessageBox.warning")
@patch("hyloa.data.session.QFileDialog.getOpenFileName", return_value=("", ""))
def test_load_session_cancelled(mock_get_open, mock_warning):
    fake_app = MagicMock()
    
    load_previous_session(fake_app)
    
    mock_warning.assert_called_once()
    mock_get_open.assert_called_once()

@patch("hyloa.data.session.QMessageBox.information")
@patch("hyloa.data.session.PlotControlWidget")
@patch("hyloa.data.session.PlotSubWindow")
@patch("hyloa.data.session.pickle.load")
@patch("builtins.open", create=True)
@patch("hyloa.data.session.QFileDialog.getOpenFileName")
def test_load_session_success(mock_get_open, mock_open, mock_pickle_load,
                               mock_plotsub, mock_plotctrl, mock_info):
    
    mock_get_open.return_value = ("dummy_path.pkl", "")
    
    
    fake_session = {
        "dataframes": ["df1", "df2"],
        "header_lines": ["h1"],
        "fit_results": {"result": 42},
        "number_plots": 2,
        "plot_widgets": {
            "0": {
                "selected_pairs": [("file.csv", "x", "y")],
                "plot_customizations": {"color": "red"}
            }
        },
        "plot_names": {
            0: "Graph 0"
        },
        "log_filename": "log.txt"
    }
    mock_pickle_load.return_value = fake_session

    # Finto app
    fake_app = MagicMock()
    fake_app.mdi_area = MagicMock()
    fake_app.figure_subwindows = {}

    load_previous_session(fake_app)

  
    mock_open.assert_called_with("dummy_path.pkl", "rb")


    assert fake_app.dataframes == ["df1", "df2"]
    assert fake_app.header_lines == ["h1"]
    assert fake_app.fit_results == {"result": 42}
    assert fake_app.number_plots == 2

    
    mock_info.assert_called_once()


@patch("hyloa.data.session.QMessageBox.critical")
@patch("hyloa.data.session.pickle.load", side_effect=Exception("fake error"))
@patch("builtins.open", create=True)
@patch("hyloa.data.session.QFileDialog.getOpenFileName")
def test_load_session_exception(mock_get_open, mock_open, mock_pickle_load, mock_critical):
    mock_get_open.return_value = ("dummy_path.pkl", "")
    fake_app = MagicMock()

    load_previous_session(fake_app)

    mock_critical.assert_called_once()



# This file is part of HYLOA - HYsteresis LOop Analyzer.
# Copyright (C) 2024 Francesco Zeno Costanzo

# HYLOA is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# HYLOA is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with HYLOA. If not, see <https://www.gnu.org/licenses/>.

"""
test error print format
"""
import pytest
from hyloa.utils.err_format import format_value_error 


@pytest.mark.parametrize("val, err, expected", [
    
    (1.234598,   0.01631,    "1.23(2)"),
    (1234.523,   12.0,       "1235(12)"),
    (2.0,        0.3,        "2.0(3)"),
    (2.0,        0.05,       "2.00(5)"),
    (12300000.0, 340000.0,   "1.23(3)e7"),
    (5.6e-05,    1.2e-06,    "5.6(1)e-5"),
    (0.001234,   5.6e-05,    "1.23(6)e-3"),
    (2000,       150,        "2000(150)"),
    (2000,       3,          "2000(3)"),
    (0.983493,   0.00021341, "9.835(2)e-1"),
    (1e-8,       3e-9,       "1.0(3)e-8"),
    (9.876e7,    2.3e6,      "9.9(2)e7"),
    (2,          7423,       "2(7423)e0")
])

def test_format_value_error(val, err, expected):
    result = format_value_error(val, err)
    assert result == expected, f"Expected {expected}, got {result}"



# This file is part of HYLOA - HYsteresis LOop Analyzer.
# Copyright (C) 2024 Francesco Zeno Costanzo

# HYLOA is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# HYLOA is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with HYLOA. If not, see <https://www.gnu.org/licenses/>.

"""
Code connected to the button to check if the installed version is
the last version aviable on GitHub
"""
import re
import requests
from packaging import version
from PyQt5.QtWidgets import QMessageBox 

import hyloa

def get_latest_version_from_init():
    '''
    Fetches the latest version of the hyloa package
    from its __init__.py file on GitHub.

    Returns
    -------
    str or None
        The latest version string if found, otherwise None.
    '''
    repo = "Francesco-Zeno-Costanzo/hyloa/main/hyloa/__init__.py"
    url  = f"https://raw.githubusercontent.com/{repo}"
    
    response = requests.get(url)
    if response.status_code == 200:
        # Extract the version using regex
        match = re.search(r'__version__\s*=\s*["\']([\d\.]+)["\']', response.text)
        if match:
            return match.group(1)
    return None

def get_local_version():
    ''' Returns the local version of the hyloa package from its __init__.py file.
    '''
    return hyloa.__version__


def is_update_available(local_ver, remote_ver):
    '''
    Checks if a newer version of the hyloa package is available.
    
    Parameters
    ----------
    local_ver : str
        The local version of the hyloa package.
    remote_ver : str
        The remote version of the hyloa package.
    
    Returns
    -------
    bool
        True if a newer version is available, False otherwise.
    '''
    return version.parse(remote_ver) > version.parse(local_ver)


def check_for_updates():
    '''
    Checks for updates to the hyloa package by comparing the local version
    with the latest version available on GitHub.
    '''
    local_ver  = get_local_version()
    remote_ver = get_latest_version_from_init()

    if remote_ver is None:
        QMessageBox.critical(None, "Error", "Unable to check for updates. Please check your internet connection")
    elif is_update_available(local_ver, remote_ver):
        QMessageBox.information(None, "Update Available",
                                f"Version {remote_ver} available (your version: {local_ver}).\n")
        
        """
        Something to aoutomatically update the package.
        For now, we just inform the user.
        """
    else:
        QMessageBox.information(None, "Already update", f"You are already using the latest version ({local_ver}).")



# This file is part of HYLOA - HYsteresis LOop Analyzer.
# Copyright (C) 2024 Francesco Zeno Costanzo

# HYLOA is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# HYLOA is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with HYLOA. If not, see <https://www.gnu.org/licenses/>.

"""
test for logger
"""
import time
import pytest
import logging
from hyloa.utils.logging_setup import setup_logging, start_logging

# Test setup_logging() to ensure it creates a file and logs messages to it
def test_setup_logging(tmp_path):
    # Define a temporary log file path
    log_file = tmp_path / "test.log"

    # Setup the logger with this path
    setup_logging(str(log_file))

    # Get the root logger and write a test message
    logger = logging.getLogger()
    logger.setLevel(logging.INFO)
    logger.info("Test message")

    # Give the file system a moment to write (sometimes needed on CI)
    time.sleep(0.1)

    # Flush all handlers to ensure writing is complete
    for handler in logger.handlers:
        handler.flush()

    # Assert the file exists and contains the test message
    assert log_file.exists(), "Log file was not created"
    content = log_file.read_text()
    assert "Test message" in content, "Log message not found in file"

    with pytest.raises(Exception) as e_info:
        setup_logging(7)


# Test start_logging() with QFileDialog mocked to return a known path
def test_start_logging_success(monkeypatch, tmp_path):
    selected_file = tmp_path / "selected.log"

    # Mock QFileDialog.getSaveFileName to return a known path
    monkeypatch.setattr(
        "PyQt5.QtWidgets.QFileDialog.getSaveFileName",
        lambda *args, **kwargs: (str(selected_file), None)
    )

    # Mock QMessageBox methods to avoid GUI interruption
    monkeypatch.setattr("PyQt5.QtWidgets.QMessageBox.information", lambda *args, **kwargs: None)
    monkeypatch.setattr("PyQt5.QtWidgets.QMessageBox.critical", lambda *args, **kwargs: None)

    # Create a fake app object with a logger attribute
    class FakeApp:
        def __init__(self):
            self.logger = None
    app = FakeApp()

    # Call start_logging, which should assign the logger and create the file
    start_logging(app)

    # Write a log message to trigger file creation
    app.logger.info("Test entry for file creation")

    time.sleep(0.1)
    for handler in app.logger.handlers:
        handler.flush()

    # Assert the log file was created and contains the test message
    assert selected_file.exists(), "Selected log file was not created"
    content = selected_file.read_text()
    assert "Test entry for file creation" in content, "Expected log content missing"

    # Test for existing logger
    start_logging(app)

    # Test invalid file
    selected_file = False
    # Mock QFileDialog.getSaveFileName to return a known path
    monkeypatch.setattr(
        "PyQt5.QtWidgets.QFileDialog.getSaveFileName",
        lambda *args, **kwargs: (selected_file, None)
    )
    app = FakeApp()
    start_logging(app)




# This file is part of HYLOA - HYsteresis LOop Analyzer.
# Copyright (C) 2024 Francesco Zeno Costanzo

# HYLOA is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# HYLOA is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with HYLOA. If not, see <https://www.gnu.org/licenses/>.

"""
Code to ensure correct print format for error
"""
import numpy as np


def format_value_error(val, err):
    '''
    A function that returns the formatting of the measurement
    with error in the form: value(error). Therefore, the error
    refers to the last digit of the value.
    Only one significant digit for the error is displayed.
    If the error is smalle than 1e-3 or bigger 1e3 the scientific
    notation will be used.
    
    Parameters
    ----------
    val : float
        central values
    err : float
        error associated to val
    
    Return
    ------
    string

    Examples
    --------
    >>> print(format_value_error(1.234598, 0.01631)
    1.23(2)
    >>> print(format_value_error(1234.523, 12.0)
    1235(12)
    >>> print(format_value_error(12300000.0 ± 340000.0)
    1.23(3)e7
    >>> print(format_value_error(5.6e-05, 1.2e-06)
    5.6(1)e-5
    '''
    
    # Division between mantissa and exponent for error
    exp  = int(np.log10(err))
    mant = err / 10**exp

    # Scientific notation for error bigger than 1e3 and smoller than 1e-3
    if abs(exp) >= 3:
        # Division between mantissa and exponent for central value
        exp_val  = int(np.log10(abs(val)) if val != 0 else 0)
        scale    = 10**exp_val
        mant_val = val / scale

        # Ensures correct formatting
        if mant_val < 1:
            exp_val  -= 1
            scale    /= 10
            mant_val *= 10   

        elif mant_val >= 10:
            exp_val  += 1
            scale    *= 10
            mant_val /= 10

        # Sacle the error
        err_scaled = err / scale
        exp_err    = int(np.log10(err_scaled))
        mant_err   = err_scaled / 10**(exp_err)

        # Ensures correct formatting
        if mant_err < 1:
            exp_err  -= 1
            # Update mantissa will be correct but not necessary
        elif mant_err >= 10:
            exp_err  += 1

        # Number of significant digits to display
        sig_dig  = 1
        # Dacimal for rounding to sig_dig digits
        decimals = max(0, -exp_err -1 + sig_dig)

        err_int  = int(round(err_scaled * 10**decimals))
        mant_val = round(mant_val, decimals)
        val_str  = f"{mant_val:.{decimals}f}"

        return f"{val_str}({err_int})e{exp_val}"

    # Ensures correct formatting
    if mant < 1.0:
        exp  -= 1 
        # Update mantissa will be correct but not necessary
    elif mant >= 10.0:
        exp  += 1

    # Number of significant digits to display
    sig_dig = 1
    # Dacimal for rounding to sig_dig digits
    decimals = max(0, -exp - 1 + sig_dig)

    err_rounded = round(err, decimals)
    val_rounded = round(val, decimals)

    err_int = int(round(err_rounded * 10**decimals))
    val_str = f"{val_rounded:.{decimals}f}"

    return f"{val_str}({err_int})"


# This file is part of HYLOA - HYsteresis LOop Analyzer.
# Copyright (C) 2024 Francesco Zeno Costanzo

# HYLOA is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# HYLOA is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with HYLOA. If not, see <https://www.gnu.org/licenses/>.

"""
test for IO
"""

import pytest
from unittest.mock import patch, mock_open, MagicMock

from hyloa.data.io import *

@pytest.fixture
def fake_app():
    return MagicMock(
        dataframes=[],
        logger=MagicMock()
    )


def test_load_files_logger_missing():
    fake_app = MagicMock()
    fake_app.logger = None

    with patch("hyloa.data.io.QMessageBox.critical") as critical_mock:
        load_files(fake_app)
        critical_mock.assert_called_once()
        assert "log" in critical_mock.call_args[0][2].lower()

@patch("hyloa.data.io.QFileDialog.getOpenFileNames")
@patch("hyloa.data.io.open", new_callable=mock_open, read_data="FieldUp\tUpRot\tUpEllipt\tIzeroUp\n")
@patch("hyloa.data.io.show_column_selection")
@patch("hyloa.data.io.detect_header_length", return_value=2)
def test_load_files_success(mock_dhl, mock_show_columns, mock_open_fn, mock_getfiles, fake_app):
    # Mock selection of the file
    mock_getfiles.return_value = (["/fake/path/data1.txt"], "")

    # First dataframe
    fake_app.dataframes = []

    # Logger 
    fake_app.logger = MagicMock()

    load_files(fake_app)

    # Verify call of QFileDialog
    mock_getfiles.assert_called_once()

    # Verify the opening of the file
    mock_open_fn.assert_called_once_with("/fake/path/data1.txt", "r", encoding='utf-8')

    # Verify the call of show_column_selection
    mock_show_columns.assert_called_once()
    args = mock_show_columns.call_args[0]
    assert args[0] == fake_app                                     # app_instance
    assert args[1] == "/fake/path/data1.txt"                       # file_path
    assert args[2] == ["FieldUp", "UpRot", "UpEllipt", "IzeroUp"]  # header


@patch("hyloa.data.io.QMessageBox.question")
@patch("hyloa.data.io.QFileDialog.getOpenFileNames")
@patch("hyloa.data.io.open", new_callable=mock_open, read_data="FieldUp\tUpRot\tUpEllipt\tIzeroUp\n")
@patch("hyloa.data.io.show_column_selection")
@patch("hyloa.data.io.detect_header_length", return_value=2)
def test_load_files_file_already_loaded(
    mock_dhl, mock_show_columns, mock_open_fn, mock_getfiles, mock_question, fake_app
    ):
    # Simulates selection of existing file
    mock_getfiles.return_value = (["/fake/path/data1.txt"], "")
    mock_question.return_value = QMessageBox.Yes 

    # Already loaded file
    df = pd.DataFrame([[0, 1, 1, 1]], columns=["FieldUp", "UpRot", "UpEllipt", "IzeroUp"])
    df.attrs["filename"] = "data1.txt"
    fake_app.dataframes = [df]
    fake_app.logger = MagicMock()

    load_files(fake_app)

    mock_getfiles.assert_called_once()
    mock_question.assert_called_once()
    mock_open_fn.assert_called_once_with("/fake/path/data1.txt", "r", encoding='utf-8')
    mock_show_columns.assert_called_once()

    # Button for no
    mock_question.return_value = QMessageBox.No

    load_files(fake_app)
    mock_getfiles.assert_called()
    mock_question.assert_called()


@patch("hyloa.data.io.QFileDialog.getOpenFileNames", return_value=([], ""))
def test_load_files_cancel_dialog(mock_getfiles, fake_app):
    load_files(fake_app)
    # No file selected
    mock_getfiles.assert_called_once()

@patch("hyloa.data.io.QFileDialog.getOpenFileNames")
@patch("hyloa.data.io.QMessageBox.critical")
@patch("hyloa.data.io.open", side_effect=Exception("Test error"))
@patch("hyloa.data.io.detect_header_length", return_value=2)
def test_load_files_open_raises(mock_dhl, mock_open_fn, mock_critical, mock_getfiles, fake_app):
    # Fake file
    mock_getfiles.return_value = (["/fake/path/data1.txt"], "")

    # mock logger
    fake_app.logger = MagicMock()

    load_files(fake_app)

    # exception handling test
    mock_critical.assert_called_once()
    error_message = mock_critical.call_args[0][2]
    assert "error" in error_message.lower()
    assert "data1.txt" in error_message


def test_detect_header_length_with_clean_header(tmp_path):
    #Create a tmp file
    content = """
        # File example
        # created: 2024
        # columns: time\tvalue
        1.0\t2.0
        2.0\t3.0
    """
    file = tmp_path / "test1.txt"
    file.write_text(content)

    result = detect_header_length(file)
    # Header with 3 rows, no empty row → data_start = 3 - 1 - 0 = 2
    assert result == 2

def test_detect_header_length_with_empty_lines(tmp_path):
    #Create a tmp file
    content = """
        # Header line
        # Unit: s\tm
            
        0.0\t0.5
        1.0\t0.6
    """
    file = tmp_path / "test2.txt"
    file.write_text(content)

    result = detect_header_length(file)
    # Header with 2 rows, 1 empty row → data_start = 3 - 1 - 1 = 1
    assert result == 1

def test_detect_header_length_with_no_numeric_data(tmp_path):
    # Create a tmp file with no numerical values
    content = """
        # just text
        text\tmoretext
        ---\t---
        nope\tagain
    """
    file = tmp_path / "test3.txt"
    file.write_text(content)

    with pytest.raises(ValueError, match="No valid data found"):
        detect_header_length(file)

def test_detect_header_length_custom_separator(tmp_path):
    # Create a tmp file csv
    content = """
        #Header A,B,C
        #units,V,A
        1.0,2.0,3.0
        4.0,5.0,6.0
    """
    file = tmp_path / "test4.csv"
    file.write_text(content)

    result = detect_header_length(file, sep=',')
    # Header with 2 rows, no empty row → data_start = 2 - 1 - 0 = 1
    assert result == 1

def test_detect_header_is_numeric_row(tmp_path):
    # Create a tmp file only numeric
    content = "1.0\t2.0\t3.0\n4.0\t5.0\t6.0\n"
    file = tmp_path / "numeric_data.txt"
    file.write_text(content)

    result = detect_header_length(file)
    # No header, no empty lines → 0 - 1 - 0 = -1
    assert result == -1

#======================================================================#
#======================================================================#
#======================================================================#


class DummyApp:
    """Fake app_instance with minimal structure."""
    def __init__(self, dataframe, header_lines):
        self.dataframes = [dataframe]
        self.header_lines = [header_lines]
        self.logger = MagicMock()


@patch("hyloa.data.io.QFileDialog.getSaveFileName")
@patch("hyloa.data.io.QMessageBox.information")
@patch("hyloa.data.io.QMessageBox.warning")
@patch("hyloa.data.io.QMessageBox.critical")
@patch("hyloa.data.io.save_header")
def test_save_to_file_success(mock_save_header, mock_critical, mock_warning, mock_info, mock_dialog, tmp_path):
    # Arrange
    test_df = pd.DataFrame({
        "A": [1.0, 2.0],
        "B": [3.0, 4.0]
    })
    header_lines = ["# test header\n"]
    app_instance = DummyApp(test_df, header_lines)

    fake_file = tmp_path / "output.txt"
    mock_dialog.return_value = (str(fake_file), "Text file (*.txt)")

    # Call function to test
    save_to_file(0, app_instance, parent_widget=None)

    # Assert all is working fine
    assert fake_file.exists()

    with open(fake_file, encoding="utf-8") as f:
        content = f.read()
        assert "1.0" in content
        assert "3.0" in content

    # Assert: save_header was called
    mock_save_header.assert_called_once_with(app_instance, header_lines, str(fake_file))
    mock_info.assert_called_once()
    mock_warning.assert_not_called()
    mock_critical.assert_not_called()


@patch("hyloa.data.io.QFileDialog.getSaveFileName")
@patch("hyloa.data.io.QMessageBox.warning")
def test_save_to_file_cancel(mock_warning, mock_dialog):
    # Arrange: simulate user cancels dialog
    mock_dialog.return_value = ("", "")
    df = pd.DataFrame({"A": [1, 2]})
    app = DummyApp(df, ["# header\n"])

    # Act
    save_to_file(0, app, parent_widget=None)

    # Assert: warning called, no file created
    mock_warning.assert_called_once()



def test_save_header_success(tmp_path):
    # Arrange
    df = pd.DataFrame({
        "col1": ["val1", "val2"],
        "col2": ["val3", np.nan],
    })
    app = DummyApp(None, None)
    file_path = tmp_path / "output.txt"

    # Call
    save_header(app, df, str(file_path))

    # Assert
    assert file_path.exists()

    with open(file_path, "r", encoding="utf-8") as f:
        lines = f.readlines()

    assert lines[0].strip() == "col1\tcol2"
    assert lines[1].strip() == "val1\tval3"
    assert lines[2].strip() == "val2"  # la col2 was NaN → must be empty

    # Logger call
    app.logger.info.assert_called_with(f"File saved successfully in: {file_path}")


def test_save_header_exception(tmp_path):
    
    app = DummyApp(None, None)
    bad_path = tmp_path / "nonexistent" / "file.txt"
    df = pd.DataFrame({"col": ["val"]})

    # Call
    save_header(app, df, str(bad_path))

    # Assert "Error while saving"
    called_msg = app.logger.info.call_args[0][0]
    assert "Error while saving" in called_msg


def test_duplicate_file_success(tmp_path):
    # create a tmp file
    original_file = tmp_path / "example.txt"
    original_file.write_text("Test content")

    expected_copy = tmp_path / "example_copy.txt"

    with patch("hyloa.data.io.QFileDialog.getOpenFileName") as mock_dialog, \
         patch("hyloa.data.io.QMessageBox.information") as mock_info, \
         patch("hyloa.data.io.QMessageBox.critical") as mock_critical :

        mock_dialog.return_value = (str(original_file), "Text Files (*.txt)")

        # Call
        duplicate_file()

        # Assert
        assert expected_copy.exists()
        assert expected_copy.read_text() == "Test content"
        mock_info.assert_called_once()
        mock_critical.assert_not_called()

def test_duplicate_file_cancel():
    with patch("hyloa.data.io.QFileDialog.getOpenFileName") as mock_dialog, \
         patch("hyloa.data.io.QMessageBox.information") as mock_info, \
         patch("hyloa.data.io.QMessageBox.critical") as mock_critical:

        mock_dialog.return_value = ("", "")  # Cancel dialog
        duplicate_file()

        mock_info.assert_not_called()
        mock_critical.assert_not_called()


# This file is part of HYLOA - HYsteresis LOop Analyzer.
# Copyright (C) 2024 Francesco Zeno Costanzo

# HYLOA is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# HYLOA is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with HYLOA. If not, see <https://www.gnu.org/licenses/>.


"""
Code to automize and incorporate in the gui a
standard correction process for hystersis loops
"""

from matplotlib.figure import Figure
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.backends.backend_qt5agg import NavigationToolbar2QT as NavigationToolbar


from PyQt5.QtWidgets import (QWidget, QHBoxLayout, QVBoxLayout, QGridLayout,
     QLabel, QLineEdit, QComboBox, QPushButton, QTextEdit, QMessageBox,
     QSizePolicy, QMdiSubWindow)
from PyQt5.QtWidgets import QScrollArea

from PyQt5.QtCore import Qt

from hyloa.data.correction import *
from hyloa.data.anisotropy import *

def correct_hysteresis_loop(app_instance):
    '''
    Opens a hysteresis loop correction window in a sub-MDI window.
    This window allows the user to set all parameters for the correction
    of a selected hysteresis loop and visualize the results.

    Help guide:

    • Select the main data file and choose the four X/Y columns that define the Up and Down branches of the loop.
    These selections must correspond to the same variables shown in the plotting window.

    • Optionally, you may select a second file where corrected data will be saved.
    The corrected columns will be written in the corresponding positions of the destination grid,
    so the layout of the two 2x2 grids matches (Up/Down x X/Y).

    • You can shift and scale the field array. Remember that the first change that is applied is the shift and then the scaling.
    These changes are not incremental, so you can safely change the value and press the button again.

    • Set the x_start / x_end limits for each branch. If the field is scaled, remember that
    the fit ranges must be chosen according to the *original*, non-scaled data; the scaling will be applied later.

    • Configure the first fit block, which is used to correct possible drifts at the end of each branch.
    You may use any polynomial or analytical function, but it is strongly recommended to avoid excessively
    high polynomial orders. The constant term MUST be the first parameter, exactly as shown in the example below.
    (e.g. 'a + b*x + c*x**2').
    To proceed with the fit of the physical quantities, you must first press the button to perform the drift correction.

    • There is also the option to replace one branch by duplicating the other. This operation can only be performed on the correct data.

    • Configure the second fit block. This works exactly like the 'Quick Fit' window and can be used
    to extract properties such as coercive field or remanence. This second fit operates on the *corrected* data.
    So in this case the fit ranges must be chosen according to the *scaled* data.
    It is recommended to use simple low-order polynomials to fit coercivity or remanence regions.
    If there's a significant discrepancy between the coercives, you can shift the field values and rerun the fit by pressing the appropriate button.
    Note that this time the shift is incremental, so a second shift will be applied to data already shifted by the first shift.

    • Cubic spline fitting and symmetrization:
    After the drift correction, you may fit each branch using a cubic B-spline.
    This step is mainly intended for data symmetrization and for the estimation of the anisotropy field.
    The smoothing parameter controls how closely the spline follows the data:
    - s = 0 corresponds to an exact interpolation of the points;
    - s > 0 allows for smoothing and helps reduce the impact of noise.
    A reasonable value depends on the data quality and on the estimated experimental error.

    
    Parameters
    ----------
    app_instance : MainApp
        The main application instance containing dataframes and logger.
    '''
    dataframes  = app_instance.dataframes
    logger      = app_instance.logger

    plot_state = {
        "flipped"   : False,
        "done_corr" : False,
        "done_spl3" : False,
        "x_up"      : None,
        "y_up"      : None,
        "x_dw"      : None,
        "y_dw"      : None,
        "x_up_corr" : None,
        "y_up_corr" : None,
        "x_dw_corr" : None,
        "y_dw_corr" : None,
        "e_up"      : None,
        "e_dw"      : None,
        "fit_hc_n"  : None,
        "fit_hc_p"  : None,
        "spline_up" : None,
        "spline_dw" : None,
        "s_data_up" : None,
        "s_data_dw" : None
    }


    if not dataframes:
        QMessageBox.critical(app_instance, "Error", "No data loaded!")
        return

    # Create main window layout
    window = QWidget()
    window.setWindowTitle("Loop Correction")
    root_layout = QHBoxLayout(window)
    window.setLayout(root_layout)

    #=========================================#
    # Control's panel on the left side        #
    #=========================================#
    
    left_widget = QWidget()
    left_layout = QVBoxLayout(left_widget)

    scroll_area = QScrollArea()
    scroll_area.setWidgetResizable(True)
    scroll_area.setWidget(left_widget)

    root_layout.addWidget(scroll_area, stretch=0)


    def show_help_dialog():
        '''
        Show a help dialog with instructions for using the correction tool.
        '''

        help_text = (
            "Help guide:\n"
            "\n"
            "• Select the main data file and choose the four X/Y columns that define the Up and Down branches of the loop.\n"
            " These selections must correspond to the same variables shown in the plotting window.\n"
            "\n"
            "• Optionally, you may select a second file where corrected data will be saved.\n"
            "The corrected columns will be written in the corresponding positions of the destination grid, "
            "so the layout of the two 2x2 grids matches (Up/Down x X/Y).\n"
            "\n"
            "• You can shift and scale the field array, remember that the first change that is applied is the shift and then the scaling. \n"
            " These changes are not incremental, so you can safely change the value and press the button again.\n"
            "\n"
            "• Set the x_start / x_end limits for each branch. If the field is scaled, remember that "
            "the fit ranges must be chosen according to the *original*, non-scaled data; the scaling will be applied later.\n"
            "\n"
            "• Configure the first fit block, which is used to correct possible drifts at the end of each branch.\n"
            "You may use any polynomial or analytical function, but it is strongly recommended to avoid excessively "
            "high polynomial orders. The constant term MUST be the first parameter, exactly as shown in the example below.\n"
            "  (e.g. 'a + b*x + c*x**2').\n"
            "To proceed with the fit of the physical quantities, you must first press the button to perform the drift correction."
            "If a file has been selected for saving, the corrected data will be saved.\n"
            "\n"
            "• There is also the option to replace one branch by duplicating the other. This operation can only be performed on the correct data.\n"
            "\n"
            "• Configure the second fit block. This works exactly like the 'Quick Fit' window and can be used "
            "to extract properties such as coercive field or remanence. This second fit operates on the *corrected* data.\n"
            "So in this case the fit ranges must be chosen according to the *scaled* data.\n"
            "It is recommended to use simple low-order polinomials to fit coercivity or remenance regions.\n"
            "If there's a significant discrepancy between the coercives, you can shift the field values and rerun the fit by pressing the appropriate button.\n"
            "Note that this time the shift is incremental, so a second shift will be applied to data already shifted by the first shift.\n"
            "\n"
            "• Cubic spline fitting and symmetrization:\n"
            "After the drift correction, you may fit each branch using a cubic B-spline. "
            "This step is mainly intended for data symmetrization and for the estimation of the anisotropy field.\n"
            "The smoothing parameter controls how closely the spline follows the data:\n"
            "  - s = 0 corresponds to an exact interpolation of the points;\n"
            "  - s > 0 allows for smoothing and helps reduce the impact of noise.\n"
            "A reasonable value depends on the data quality and on the estimated experimental error.\n"
            "If you press the button to symmetrize the data and a file has been selected for saving, the symmetrized data will be saved.\n"
            "\n"
            "• By moving the mouse over the various boxes, a tooltip appears with information relating to them."
        )

        msg = QMessageBox(window)
        msg.setWindowTitle("Correction Guide")
        msg.setIcon(QMessageBox.Information)

        text = QTextEdit()
        text.setPlainText(help_text)
        text.setReadOnly(True)
        text.setMinimumSize(700, 500)

        layout = msg.layout()
        layout.addWidget(text, 0, 0, 1, layout.columnCount())

        msg.exec_()

    

    help_button = QPushButton("Help")
    help_button.setToolTip("Press for a quick user guide")
    help_button.clicked.connect(show_help_dialog)
    left_layout.addWidget(help_button, alignment=Qt.AlignLeft)

    #===============================================#
    # Load data                                     #
    #===============================================#

    # Selection of the file to process
    left_layout.addWidget(QLabel("Select data file (source):"))
    file_combo = QComboBox()
    file_combo.addItems([f"File {i+1}" for i in range(len(dataframes))])
    file_combo.setToolTip("File from which load data")
    left_layout.addWidget(file_combo)

    # Selecetion of the columns that contain the data
    box_data = QGridLayout()
    left_layout.addLayout(box_data)

    box_data.addWidget(QLabel("up: "), 0, 0)
    x_up_combo = QComboBox(); box_data.addWidget(x_up_combo, 0, 1)
    y_up_combo = QComboBox(); box_data.addWidget(y_up_combo, 0, 2)
    x_up_combo.setToolTip("x data for up branch")
    y_up_combo.setToolTip("y data for up branch")

    box_data.addWidget(QLabel("down: "), 1, 0)
    x_down_combo = QComboBox(); box_data.addWidget(x_down_combo, 1, 1)
    y_down_combo = QComboBox(); box_data.addWidget(y_down_combo, 1, 2)
    x_down_combo.setToolTip("x data for down branch")
    y_down_combo.setToolTip("y data for down branch")

    def update_columns_from_selected_file():
        ''' Update the column selection combos based on the selected source file.
        '''
        idx  = file_combo.currentIndex()
        cols = [c for c in list(dataframes[idx].columns) if str(c) != '']

        for combo in (x_up_combo, y_up_combo, x_down_combo, y_down_combo):
            # Block signals to avoid too rapid changes
            combo.blockSignals(True)
            combo.clear()
            combo.addItems(cols)
            combo.setCurrentIndex(0)
            combo.blockSignals(False)

    file_combo.currentIndexChanged.connect(update_columns_from_selected_file)
    update_columns_from_selected_file()

    #===============================================#
    # Store data                                    #
    #===============================================#
    
    # Selection of the file to save corrected columns
    left_layout.addWidget(QLabel("Optional: file to save corrected data (use same order):"))
    save_file_combo = QComboBox()
    save_file_combo.addItem("No save")
    save_file_combo.addItems([f"File {i+1}" for i in range(len(dataframes))])
    save_file_combo.setToolTip("File to store data. Suggest to use duplicate file to create it.")
    left_layout.addWidget(save_file_combo)

    # Selection of destination columns
    dest_box = QGridLayout()
    left_layout.addLayout(dest_box)

    dest_box.addWidget(QLabel("Save up:"), 0, 0)
    x_up_dest = QComboBox(); dest_box.addWidget(x_up_dest, 0, 1)
    y_up_dest = QComboBox(); dest_box.addWidget(y_up_dest, 0, 2)
    x_up_dest.setToolTip("In this variable the data on the x of the up branch will be stored.")
    y_up_dest.setToolTip("In this variable the data on the y of the up branch will be stored.")


    dest_box.addWidget(QLabel("Save down:"), 1, 0)
    x_dw_dest = QComboBox(); dest_box.addWidget(x_dw_dest, 1, 1)
    y_dw_dest = QComboBox(); dest_box.addWidget(y_dw_dest, 1, 2)
    x_dw_dest.setToolTip("In this variable the data on the x of the down branch will be stored.")
    y_dw_dest.setToolTip("In this variable the data on the y of the down branch will be stored.")


    def update_dest_columns():
        ''' Update the column selection combos based on the selected save file.
        '''
        save_choice = save_file_combo.currentIndex()

        if save_choice == 0:
            # No save → empty combos
            for combo in (x_up_dest, y_up_dest, x_dw_dest, y_dw_dest):
                combo.clear()
            return

        dest_idx = save_choice - 1  
        cols = list(dataframes[dest_idx].columns)

        for combo in (x_up_dest, y_up_dest, x_dw_dest, y_dw_dest):
            combo.clear()
            combo.addItems(cols)

    save_file_combo.currentIndexChanged.connect(update_dest_columns)

    #===============================================#
    # Set parameters for the field corrections      #
    #===============================================#

    left_layout.addWidget(QLabel("-------- Preliminary field changes --------"))

    st_grid = QGridLayout()
    left_layout.addLayout(st_grid)
    st_grid.addWidget(QLabel("Shift (i.e. H = H - shift)"), 0, 0)
    st_grid.addWidget(QLabel("Scale (i.e. H = H * scale)"), 0, 1)
    st_grid.addWidget(QLabel("Revert a branch"),            0, 2)

    
    field_shift_edit = QLineEdit("0")
    field_shift_edit.setToolTip("Both field arrays will be shifted by this value. "
                                "The shift is applied before scaling.")
    st_grid.addWidget(field_shift_edit, 1, 0)

    field_scale_edit = QLineEdit("1")
    field_scale_edit.setToolTip("Both field arrays will be scaled by this value."
                                "The scaling will be applied after the shift")
    st_grid.addWidget(field_scale_edit, 1, 1)

    flip_btn = QPushButton("Check Symmetry")
    flip_btn.setToolTip("Revert up branch to visually check the symmetry.")
    st_grid.addWidget(flip_btn, 1, 2)

    #===============================================#
    # Set parameters for the mag corrections        #
    #===============================================#

    left_layout.addWidget(QLabel("-------- Fit for tail correction --------"))

    limits_grid = QGridLayout()
    left_layout.addLayout(limits_grid)

    x_start_n_edit = QLineEdit("-4000")
    x_end_n_edit   = QLineEdit("-400")

    x_start_n_edit.setToolTip("Left limit for fit in the negative field saturation zone.\n"
                              "If greater than the largest value of the fields, "
                              "data will be taken starting from the largest existing value.")
    x_end_n_edit.setToolTip("Right limit for fit in the negative field saturation zone.\n"
                            "Choose carefully based on the data.")
    
    limits_grid.addWidget(QLabel("x_start_neg:"), 0, 0)
    limits_grid.addWidget(x_start_n_edit,         0, 1)
    limits_grid.addWidget(QLabel("x_end_neg:"),   0, 2)
    limits_grid.addWidget(x_end_n_edit,           0, 3)

    x_start_p_edit = QLineEdit("400")
    x_end_p_edit   = QLineEdit("4000")

    x_start_p_edit.setToolTip("Left limit for fit in the positive field saturation zone.\n"
                              "Choose carefully based on the data.")
    x_end_p_edit.setToolTip("Right limit for fit in the positive field saturation zone.\n"
                            "If greater than the largest value of the fields, "
                            "data will be taken starting from the largest existing value.")
                            

    limits_grid.addWidget(QLabel("x_start_pos:"), 1, 0)
    limits_grid.addWidget(x_start_p_edit,         1, 1)
    limits_grid.addWidget(QLabel("x_end_pos:"),   1, 2)
    limits_grid.addWidget(x_end_p_edit,           1, 3)

    box_fit_params = QGridLayout()
    left_layout.addLayout(box_fit_params)
    box_fit_params.addWidget(QLabel("Parameter names (e.g. a,b):"), 0, 0)
    box_fit_params.addWidget(QLabel("Initial fit values (e.g. 1,1):"), 0, 1)

    tail_params_edit = QLineEdit("a,b")
    box_fit_params.addWidget(tail_params_edit, 1, 0)

    tail_initials_edit = QLineEdit("1,1")
    box_fit_params.addWidget(tail_initials_edit, 1, 1)

    box_fit_params.addWidget(QLabel("Function (e.g. a + b*x) :"), 2, 0)
    tail_function_edit = QLineEdit("a + b*x")
    box_fit_params.addWidget(tail_function_edit, 2, 1)

    #===============================================#
    # Correction buttons                            #
    #===============================================#

    corr_btn_box = QGridLayout()
    left_layout.addLayout(corr_btn_box)
    
    run_button = QPushButton("Apply drift correction")
    run_button.setToolTip("Remove drift in the saturation zone")
    corr_btn_box.addWidget(run_button, 0, 0)

    del_cp_btn = QPushButton("Remove corrections")
    del_cp_btn.setToolTip("Remove from all done corrections from plot and delete them")
    corr_btn_box.addWidget(del_cp_btn, 0, 1)

    del_od_btn = QPushButton("Remove original data")
    del_od_btn.setToolTip("Remove original data from plot")
    corr_btn_box.addWidget(del_od_btn, 0, 2)

    box_options = QGridLayout()
    left_layout.addLayout(box_options)
    # Select option to duplicate a branch
    double_branch = QComboBox()
    double_branch.addItem("No")
    double_branch.addItem("Up")
    double_branch.addItem("Down")
    box_options.addWidget(QLabel("Duplicate branch: "), 0, 0)
    box_options.addWidget(double_branch, 0, 1)
    double_btn = QPushButton("Duplicate")
    box_options.addWidget(double_btn, 0, 2)

    #===============================================#
    # Data selection for parameter estimations      #
    #===============================================#

    data_used = QGridLayout()
    left_layout.addLayout(data_used)
    # Select option to duplicate a branch
    data_sel = QComboBox()
    data_sel.addItem("Original")
    data_sel.addItem("Corrected")
    data_used.addWidget(QLabel("Select data for fit: "), 0, 0)
    data_used.addWidget(data_sel, 0, 1)

    #===============================================#
    # Set parameters for coercivity computation     #
    #===============================================#

    left_layout.addWidget(QLabel("-------- Fit for coercivity/remenance estimation --------"))

    limits_grid_1 = QGridLayout()
    left_layout.addLayout(limits_grid_1)

    x_start_up_hc_edit = QLineEdit("100")
    x_end_up_hc_edit   = QLineEdit("300")

    x_start_up_hc_edit.setToolTip("Left limit for fit for up branch.\n"
                                  "Choose carefully based on the data.")
    x_end_up_hc_edit.setToolTip("Right limit for fit for up branch.\n"
                                "Choose carefully based on the data.")
    
    limits_grid_1.addWidget(QLabel("x_start_up:"), 0, 0)
    limits_grid_1.addWidget(x_start_up_hc_edit,    0, 1)
    limits_grid_1.addWidget(QLabel("x_end_up:"),   0, 2)
    limits_grid_1.addWidget(x_end_up_hc_edit,      0, 3)

    x_start_dw_hc_edit = QLineEdit("-300")
    x_end_dw_hc_edit   = QLineEdit("-100")

    x_start_dw_hc_edit.setToolTip("Left limit for fit for down branch.\n"
                                  "Choose carefully based on the data.")
    x_end_dw_hc_edit.setToolTip("Right limit for fit for down branch.\n"
                                "Choose carefully based on the data.")

    limits_grid_1.addWidget(QLabel("x_start_down:"), 1, 0)
    limits_grid_1.addWidget(x_start_dw_hc_edit,      1, 1)
    limits_grid_1.addWidget(QLabel("x_end_down:"),   1, 2)
    limits_grid_1.addWidget(x_end_dw_hc_edit,        1, 3)

    box_fit_params_1 = QGridLayout()
    left_layout.addLayout(box_fit_params_1)
    box_fit_params_1.addWidget(QLabel("Parameter names (e.g. a,b):"), 0, 0)
    box_fit_params_1.addWidget(QLabel("Initial fit values (e.g. 1,1):"),  0, 1)

    hc_params_edit = QLineEdit("s,hc")
    box_fit_params_1.addWidget(hc_params_edit, 1, 0)

    hc_initials_edit = QLineEdit("1,1")
    box_fit_params_1.addWidget(hc_initials_edit, 1, 1)

    fit_btn_box = QGridLayout()
    left_layout.addLayout(fit_btn_box)

    fit_btn_box.addWidget(QLabel("Function (e.g. s*(x - hc) ):"), 0, 0)
    hc_function_edit = QLineEdit("s*(x - hc)")
    fit_btn_box.addWidget(hc_function_edit, 0, 1)

    fit_btn = QPushButton("fit")
    fit_btn_box.addWidget(fit_btn, 0, 2)

    #===============================================#
    # Correction buttons                            #
    #===============================================#

    left_layout.addWidget(QLabel("-------- Final correction of the fields a posterior --------"))

    fit_btn_box = QGridLayout()
    left_layout.addLayout(fit_btn_box)
    
    fit_btn_box.addWidget(QLabel("Shift (i.e. H = H - shift)"), 0, 1)
    
    field_shift_pc_edit = QLineEdit("0")
    field_shift_pc_edit.setToolTip("This shift will be applied to the correct data.")
    fit_btn_box.addWidget(field_shift_pc_edit, 0, 2)

    field_shift_btn = QPushButton("Apply shift and fit again")
    field_shift_btn.setToolTip("Fit again corrected data after the shift.")
    fit_btn_box.addWidget(field_shift_btn, 0, 3)


    #===============================================#
    # Anisotropy Field estimation                   #
    #===============================================#

    left_layout.addWidget(QLabel("-------- Symmetrization and estimation of the anisotropy field --------"))

    box_spline = QGridLayout()
    left_layout.addWidget(QLabel("Smothing factor for up and down brach cubic spline fitting:"))

    left_layout.addLayout(box_spline)

    box_spline.addWidget(QLabel("Up branch:"), 0, 0)
    smooth_up_edit = QLineEdit("0.01")
    smooth_up_edit.setToolTip("Smoothing Bsline parameter for the up branch.\n"
                              "After correcting the data, the value is automatically changed, "
                              "providing a more accurate estimate based on the data error. \n"
                              "If 0 normal interpolation.")
    box_spline.addWidget(smooth_up_edit, 0, 1)

    box_spline.addWidget(QLabel("Dw branch:"), 0, 2)
    smooth_dw_edit = QLineEdit("0.01")
    smooth_up_edit.setToolTip("Smoothing Bspline parameter for the down branch.\n"
                              "After correcting the data, the value is automatically changed, "
                              "providing a more accurate estimate based on the data error.\n"
                              "If 0, normal interpolation.")
    box_spline.addWidget(smooth_dw_edit, 0, 3)

    spl3_btn_box = QGridLayout()
    left_layout.addLayout(spl3_btn_box)
    
    spl3_btn = QPushButton("Create spline")
    spl3_btn.setToolTip("Fit data with a cubic spline.")
    spl3_btn_box.addWidget(spl3_btn, 0, 0)

    sym_btn = QPushButton("Symmetrize loop")
    sym_btn.setToolTip("Create a new loop strating form the preconstructed spline.")
    spl3_btn_box.addWidget(sym_btn, 0, 1)

    del_sym_btn = QPushButton("Remove sym loop")
    del_sym_btn.setToolTip("Remove new loop form the plot and delete them.")
    spl3_btn_box.addWidget(del_sym_btn, 0, 2)

    hk_box = QGridLayout()
    left_layout.addLayout(hk_box)

    hk_box.addWidget(QLabel("Closure threshold:"), 0, 0)
    hk_thr_edit = QLineEdit("0.02")
    hk_thr_edit.setToolTip("Threshold below which the relative difference of the branches must be to find the closing point.")
    hk_box.addWidget(hk_thr_edit, 0, 1)
    hk_btn = QPushButton("Compute anisotropy field")
    hk_box.addWidget(hk_btn, 0, 2)
    

    #===============================================#

    left_layout.addStretch(1)
   
    #===============================================#
    # ----------- RIGHT: plot + results ----------- #
    #===============================================#

    right_layout = QVBoxLayout()
    root_layout.addLayout(right_layout, stretch=1)

    # Matplotlib figure & canvas
    fig = Figure(figsize=(5,4))
    canvas = FigureCanvas(fig)

    # Toolbar Matplotlib (zoom, pan, save, coordinate)
    toolbar = NavigationToolbar(canvas, window)
    right_layout.addWidget(toolbar)

    ax = fig.add_subplot(111)
    ax.set_title("Data preview")
    ax.set_xlabel("H [Oe]", fontsize=15)
    ax.set_ylabel(r"M/M$_{sat}$", fontsize=15)
    canvas.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
    right_layout.addWidget(canvas, stretch=3)

    # Output box
    output_box = QTextEdit()
    output_box.setReadOnly(True)
    output_box.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
    output_box.setMaximumHeight(140)

    right_layout.addWidget(output_box, stretch=1)

    def draw_plot():
        ax.clear()

        # flip factor (non altera i dati direttamente)
        mul = -1 if plot_state["flipped"] else 1

        #==========================================================#
        # Raw data                                                 #
        #==========================================================#
        if plot_state["x_up"] is not None and plot_state["y_up"] is not None:

            ax.plot(plot_state["x_up"] * mul, plot_state["y_up"] * mul, 'k.-',
                    label="Up raw", alpha=0.5 if plot_state["done_corr"] else 1)
            
        if plot_state["x_dw"] is not None and plot_state["y_dw"] is not None:

            ax.plot(plot_state["x_dw"], plot_state["y_dw"], 'k.-', 
                    label="Down raw", alpha=0.5 if plot_state["done_corr"] else 1)

        #==========================================================#
        # Corrected data                                           #    
        #==========================================================#
        if plot_state.get("x_up_corr") is not None:

            ax.errorbar(
                plot_state["x_up_corr"] * mul, plot_state["y_up_corr"] * mul,
                yerr=plot_state["e_up"], linestyle='-', 
                fmt='.', color='r', label="Up corrected",
                alpha=0.5 if plot_state["done_spl3"] else 1
            )

        if plot_state.get("x_dw_corr") is not None:

            ax.errorbar(
                plot_state["x_dw_corr"], plot_state["y_dw_corr"],
                yerr=plot_state["e_dw"], linestyle='-', 
                fmt='.', color='r', label="Down corrected",
                alpha=0.5 if plot_state["done_spl3"] else 1
            )

        #==========================================================#
        # Fit lines                                                #
        #==========================================================#
        if plot_state.get("fit_hc_n") is not None:
            ax.plot(*plot_state["fit_hc_n"], 'b--', label="HC neg fit")

        if plot_state.get("fit_hc_p") is not None:
            ax.plot(*plot_state["fit_hc_p"], 'b--', label="HC pos fit")

        #==========================================================#
        # Spline lines                                             #
        #==========================================================#
        if plot_state.get("spline_up") is not None:
            ax.plot(*plot_state["spline_up"], '-', color='navy', label="spl3 Up")

        if plot_state.get("spline_dw") is not None:
            ax.plot(*plot_state["spline_dw"], '-', color='navy', label="spl3 Dw")

        #==========================================================#
        # Symmetrizied data                                        #
        #==========================================================#
        if plot_state.get("s_data_up") is not None:
            ax.plot(*plot_state["s_data_up"], 'k.-', label="sym Up")

        if plot_state.get("s_data_dw") is not None:
            ax.plot(*plot_state["s_data_dw"], 'k.-', label="sym Dw")

        #==========================================================#
        ax.axhline(0, color='gray', linestyle='--', linewidth=1)
        ax.axvline(0, color='gray', linestyle='--', linewidth=1)

        if ax.get_legend_handles_labels()[1]:
            ax.legend()

        ax.set_xlabel("H [Oe]", fontsize=15)
        ax.set_ylabel("M/M$_{sat}$", fontsize=15)

        fig.tight_layout()
        canvas.draw()


    #================================================#
    # Helper function for plot and preview refresh   #
    #================================================#
    def get_preview_data():
        ''' Get initial data for preview plot
        '''
        idx = file_combo.currentIndex()
        df  = dataframes[idx]

        x_up = df[x_up_combo.currentText()].astype(float).values
        y_up = df[y_up_combo.currentText()].astype(float).values
        x_dw = df[x_down_combo.currentText()].astype(float).values
        y_dw = df[y_down_combo.currentText()].astype(float).values

        return x_up, y_up, x_dw, y_dw

    def refresh_preview(x_up, y_up, x_dw, y_dw):
        ''' Draw selected data on canvas
        '''

        try:
            plot_state.update({
                "x_up": x_up,
                "y_up": y_up,
                "x_dw": x_dw,
                "y_dw": y_dw,
            })
            draw_plot()

        except Exception as e:
            QMessageBox.critical("Error refreshing preview: %s", e)


    # Connect changes to update preview
    for cb in (x_up_combo, y_up_combo, x_down_combo, y_down_combo, file_combo):
        try:
            cb.currentIndexChanged.connect(
                lambda : refresh_preview(*get_preview_data())
            )
        except Exception:
            pass

    #================================================#
    # Button connections                             #
    #================================================#
    
    # Connect run button
    run_button.clicked.connect(lambda: perform_correction(
            file_combo, save_file_combo,
            x_up_combo, y_up_combo, x_down_combo, y_down_combo,
            field_shift_edit, field_scale_edit,
            x_start_n_edit, x_end_n_edit, x_start_p_edit, x_end_p_edit,
            tail_params_edit, tail_function_edit,
            x_up_dest, y_up_dest, x_dw_dest, y_dw_dest,
            smooth_up_edit, smooth_dw_edit,
            dataframes, logger, plot_state, draw_plot,
            window, 
        )
    )

    flip_btn.clicked.connect(lambda: flip(
            plot_state, window, draw_plot
        )
    ) 

    del_cp_btn.clicked.connect(lambda: change_ps(
            plot_state, window, draw_plot, mode="cp"
        )
    )

    del_od_btn.clicked.connect(lambda:change_ps(
            plot_state, window, draw_plot, mode="od"
        )
    )

    fit_btn.clicked.connect(lambda : fit_data(
            file_combo,
            x_up_combo, y_up_combo, x_down_combo, y_down_combo, data_sel,
            x_start_up_hc_edit, x_end_up_hc_edit, x_start_dw_hc_edit, x_end_dw_hc_edit,
            hc_params_edit, hc_function_edit, logger, plot_state, draw_plot,
            output_box, window
        )
    )

    double_btn.clicked.connect(lambda : flip_data(
            file_combo,
            x_up_combo, y_up_combo, x_down_combo, y_down_combo, data_sel,
            double_branch, plot_state,
            window, logger, draw_plot
        )
    )
    
    field_shift_btn.clicked.connect(lambda: apply_shift(
            x_up_dest, y_up_dest, x_dw_dest, y_dw_dest, dataframes, save_file_combo, data_sel,
            field_shift_pc_edit, plot_state, window, fit_data, args=(
                file_combo,
                x_up_combo, y_up_combo, x_down_combo, y_down_combo, data_sel,
                x_start_up_hc_edit, x_end_up_hc_edit, x_start_dw_hc_edit, x_end_dw_hc_edit,
                hc_params_edit, hc_function_edit, logger, plot_state, draw_plot,
                output_box, window
            ),
            logger=logger)
    )

    spl3_btn.clicked.connect(lambda: compute_b_spline(
            file_combo, x_up_combo, y_up_combo, x_down_combo, y_down_combo, data_sel,
            smooth_up_edit, smooth_dw_edit,
            plot_state, logger, window, draw_plot

        )
    )

    sym_btn.clicked.connect(lambda: symmetrize(
            file_combo, save_file_combo,
            x_up_combo, y_up_combo, x_down_combo, y_down_combo, data_sel,
            x_up_dest,  y_up_dest,  x_dw_dest,    y_dw_dest,
            dataframes, logger, plot_state, draw_plot,
            window, 
        )
    )

    del_sym_btn.clicked.connect(lambda:change_ps(
            plot_state, window, draw_plot, mode="sym"
        )
    )

    hk_btn.clicked.connect(lambda: compute_Hk(
            file_combo, x_up_combo, y_up_combo, x_down_combo, y_down_combo,
            hk_thr_edit, plot_state, logger, window, output_box
        )
    )
        
    # Sub-window for fitting panel
    sub = QMdiSubWindow()
    sub.setWidget(window)
    sub.setWindowTitle("Loop Correction")
    #sub.resize(1200, 900)
    sub.resize(window.sizeHint())

    app_instance.mdi_area.addSubWindow(sub)
    sub.show()


# This file is part of HYLOA - HYsteresis LOop Analyzer.
# Copyright (C) 2024 Francesco Zeno Costanzo

# HYLOA is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# HYLOA is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with HYLOA. If not, see <https://www.gnu.org/licenses/>.

"""
Test entry point
"""
import time
import pytest
from PyQt5.QtGui import QPixmap
from PyQt5.QtCore import Qt

from hyloa.main import *


def test_splash_creation(qtbot):
    pixmap = QPixmap(100, 100)
    splash = Splash(pixmap)

    qtbot.addWidget(splash)

    assert splash.windowFlags() & Qt.FramelessWindowHint
    assert splash.progress.value() == 0


def test_progress_update(qtbot):
    pixmap = QPixmap(100, 100)
    splash = Splash(pixmap)

    qtbot.addWidget(splash)

    splash.set_progress(50)
    assert splash.progress.value() == 50

    splash.set_progress(100)
    assert splash.progress.value() == 100


def test_progress_format(qtbot):
    pixmap = QPixmap(100, 100)
    splash = Splash(pixmap)

    qtbot.addWidget(splash)

    assert splash.progress.isTextVisible()
    assert splash.progress.format() == "%p%"



def test_remaining_time_positive():
    start = time.monotonic()
    time.sleep(0.1)

    remaining = compute_remaining_time(start, 1.0)

    assert remaining > 0
    assert remaining <= 1000


def test_remaining_time_zero_when_expired():
    start = time.monotonic()
    time.sleep(0.2)

    remaining = compute_remaining_time(start, 0.1)

    assert remaining == 0



# This file is part of HYLOA - HYsteresis LOop Analyzer.
# Copyright (C) 2024 Francesco Zeno Costanzo

# HYLOA is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# HYLOA is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with HYLOA. If not, see <https://www.gnu.org/licenses/>.


"""
Code to handle data input and output, i.e. loading and saving data
"""
import os
import shutil
import numpy as np
import pandas as pd

from PyQt5.QtWidgets import (
    QFileDialog, QMessageBox, QWidget, QHBoxLayout, QLabel,
    QPushButton, QCheckBox, QLineEdit, QScrollArea, QTableWidget,
    QTableWidgetItem, QDialog, QVBoxLayout, QComboBox,QListWidget
)
from PyQt5.QtCore import Qt


#==============================================================================================#
# File upload functions                                                                        #
#==============================================================================================#

def load_files(app_instance):
    '''
    Function to upload a file chosen by the user.
    If the file has an header, this will be preserved when the data is
    saved to a new file.

    Parameters
    ----------
    app_instance : instance of MainApp from main_window.py
    '''
    if app_instance.logger is None:
        QMessageBox.critical(None, "Error", "Cannot start analysis without starting log")
        return

    file_paths, _ = QFileDialog.getOpenFileNames(
        None,
        "Select a file",
        "",
        "Text Files (*.txt);;All Files (*)"
    )

    if not file_paths:
        return

    for file_path in file_paths:

        filename = os.path.basename(file_path)

        # Check if the file is already loaded
        existing_names = [df.attrs.get("filename", "") for df in app_instance.dataframes]

        if filename in existing_names:
            reply = QMessageBox.question(
                None,
                "File already uploaded",
                f"The file '{filename}' has already been uploaded.\nDo you want to overwrite it?",
                QMessageBox.Yes | QMessageBox.No
            )
            if reply == QMessageBox.No:
                continue  # Skip this file
            else:
                index_to_replace = existing_names.index(filename)
        else:
            index_to_replace = None  # new file

        try:
            if detect_header_length(file_path) == -1:
                data   = np.loadtxt(file_path, max_rows=1)
                n_col  = data.size
                header = [f"col_{i}" for i in range(n_col)]
            else:
                with open(file_path, "r", encoding='utf-8') as f:
                    header = f.readline().strip().split("\t")

            app_instance.logger.info(f"Apertura file {file_path}")
            show_column_selection(app_instance, file_path, header, index_to_replace)

        except Exception as e:
            QMessageBox.critical(None, "Error", f"Error loading file: {file_path}\n{e}")

#==============================================================================================#

def detect_header_length(file_path, sep='\t'):
    '''
    Function to compute the length of the header and therefore,
    the number of rows to exclude from the dataframe to obtain a
    dataframe that has for each column only the data of interest.

    Parameters
    ----------
    file_path : string
        path of the file to read
    sep : string
        separetor, optional, default a tabulation

    Return
    ------
    data_start : int
        number of the not empty lines of the file's header
    '''
    with open(file_path, "r", encoding="utf-8") as f:
        lines = f.readlines()

    data_start  = None
    empty_lines = 0
    
    def is_float(s):
        try:
            float(s)
            return True
        except Exception:
            return False

    # Find the first row with no letter
    for i, line in enumerate(lines):
        parts         = [x.strip() for x in line.strip().split(sep)]
        numeric_parts = [x for x in parts if is_float(x)]
        
        if parts[0] == '':
            empty_lines += 1
        
        if len(numeric_parts) == len(parts) and len(parts) > 1:
            data_start = i
            break

    if data_start is None:
        raise ValueError("No valid data found.")
    
    data_start = data_start - 1 - empty_lines
    
    return data_start

#==============================================================================================#

def show_column_selection(app_instance, file_path, header, index_to_replace=None):
    '''
    Dialog window to select columns to load.

    Parameters
    ----------
    app_instance : instance of MainApp
    file_path : string
        path of the file to read
    header : list of str
        header of the file
    '''
    # Create a dialog window
    dialog = QWidget()
    dialog.setWindowTitle(f"Select Columns: {os.path.basename(file_path)}")
    dialog.setGeometry(100, 100, 900, 750)
    main_layout = QVBoxLayout(dialog)

    # Instructions
    instructions = QLabel(
        "Select columns to load and name them. If the name is empty, the default will be used.\n"
        "Scroll down to view data."
    )
    instructions.setWordWrap(True)
    main_layout.addWidget(instructions)

    selected_columns = {}
    custom_names     = {}

    header_length = detect_header_length(file_path)

    # Preview Data Table
    if header_length > 0:
        all_df = pd.read_csv(file_path, sep="\t").drop(list(range(header_length)))
        
    elif header_length == -1:
        data     = np.loadtxt(file_path)
        _, n_col = data.shape
        col      = [f"col_{i}" for i in range(n_col)]
        all_df   = pd.DataFrame(data, columns=col)
    else:
        all_df = pd.read_csv(file_path, sep="\t")

    table = QTableWidget()
    table.setRowCount(len(all_df))
    table.setColumnCount(len(all_df.columns))
    table.setHorizontalHeaderLabels(list(all_df.columns))

    for i in range(len(all_df)):
        for j, col in enumerate(all_df.columns):
            item = QTableWidgetItem(str(all_df.iloc[i, j]))
            item.setFlags(item.flags() ^ Qt.ItemIsEditable)
            table.setItem(i, j, item)

    table.setFixedHeight(250)
    main_layout.addWidget(table)

    # Column selection section (in scroll area)
    scroll = QScrollArea()
    scroll.setWidgetResizable(True)
    scroll_content = QWidget()
    scroll_layout = QVBoxLayout(scroll_content)

    for i, col_name in enumerate(header):
        box = QHBoxLayout()
        checkbox = QCheckBox(col_name)
        checkbox.setChecked(True)
        selected_columns[i] = checkbox
        custom_names[i] = QLineEdit()
        custom_names[i].setPlaceholderText("Custom column name")
        box.addWidget(checkbox)
        box.addWidget(custom_names[i])
        scroll_layout.addLayout(box)

    scroll_content.setLayout(scroll_layout)
    scroll.setWidget(scroll_content)
    main_layout.addWidget(scroll)

    # Confirm button
    def submit_selection():
        try:
            columns_to_load = [header[i] for i in range(len(header)) if selected_columns[i].isChecked()]
            column_names = [
                custom_names[i].text() or f"{os.path.splitext(os.path.basename(file_path))[0]}_{header[i]}"
                for i in range(len(header)) if selected_columns[i].isChecked()
            ]

            header_length = detect_header_length(file_path)

            if header_length >= 0:
                df_header = pd.read_csv(file_path, sep="\t", nrows=header_length)
                df_data   = pd.read_csv(file_path, sep="\t", usecols=columns_to_load)

                df_data.columns = column_names

            elif header_length == -1:
                df_header = header
                data      = np.loadtxt(file_path)
                df_data   = pd.DataFrame(data, columns=column_names)
            
            if header_length >= 0 :
                df_data = df_data.drop(list(range(header_length)))

            app_instance.logger.info(f"From: {file_path}, load: {columns_to_load}")

            df_data.attrs["filename"] = os.path.basename(file_path)

            if index_to_replace is not None:
                app_instance.dataframes[index_to_replace]   = df_data
                app_instance.header_lines[index_to_replace] = df_header
                app_instance.logger.info(f"File '{file_path}' overwrite in position {index_to_replace}")
            else:
                app_instance.dataframes.append(df_data)
                app_instance.header_lines.append(df_header)
                app_instance.logger.info(f"File '{file_path}' added")

            QMessageBox.information(dialog, "Success", f"Data loaded form {file_path}!")
            app_instance.refresh_shell_variables()
            dialog.close()
        except Exception as e:
            QMessageBox.critical(dialog, "Error", f"Error while loading:\n{e}")

    confirm_button = QPushButton("Load")
    confirm_button.clicked.connect(submit_selection)
    main_layout.addWidget(confirm_button)

    dialog.show()

#==============================================================================================#
# Functions to save modified data                                                              #
#==============================================================================================#

def save_header(app_instance, df, file_path):
    '''
    Save the header in the final text file,
    to provide compatibility in case you want to
    reopen it for further analysis.

    Parameters
    ----------
    app_instance : MainApp object
        instance of MainApp from main_window.py
    df : pandas dataframe
        dataframe containing only the header of the data file
    file_path : string
        path of the file to save
    '''
    try:
        with open(file_path, "w", encoding='utf-8') as f:
            # Write the names of the columns
            f.write("\t".join(df.columns) + "\n")
            
            # Writes every row of the DataFrame, ignoring NaNs
            for _, row in df.iterrows():
                line = "\t".join(row.astype(str).fillna("").replace("nan", "").values)
                f.write(line.strip() + "\n")
        app_instance.logger.info(f"File saved successfully in: {file_path}")

    except Exception as e:
        app_instance.logger.info(f"Error while saving: {e}")

#==============================================================================================#

def save_modified_data(app_instance, parent_widget):
    ''' 
    Allows you to choose the file and data to save, with the related headers.

    Parameters
    ----------
    app_instance : MainApp object
        instance of MainApp from main_window.py
    parent_widget : QWidget
        parent widget for the dialog
    '''

    dataframes = app_instance.dataframes

    if not dataframes:
        QMessageBox.critical(parent_widget, "Error", "No data aviable.")
        return
    
    instructions = QLabel(
        "Select the files containing the data you want to save.\n"
        "You can see the available columns for each selected file."
    )
    instructions.setWordWrap(True)
    
    dialog = QDialog(parent_widget)
    dialog.setWindowTitle("Select the file to save")
    dialog.setGeometry(100, 100, 500, 300)

    layout = QVBoxLayout(dialog)
    layout.addWidget(instructions)
    layout.addWidget(QLabel("Select the file to save:"))

    combo = QComboBox()
    for i in range(len(dataframes)):
        combo.addItem(f"File {i + 1}")
    layout.addWidget(combo)

    layout.addWidget(QLabel("Columns of the selected file:"))
    column_list = QListWidget()
    layout.addWidget(column_list)

    def update_column_list(index):
        ''' Function to update the list of the columns
        '''
        column_list.clear()
        if index < len(dataframes):
            columns = dataframes[index].columns
            for col in columns:
                column_list.addItem(str(col))

    # Connect file selection and columns list
    combo.currentIndexChanged.connect(update_column_list)
    
    # Initialize the list
    update_column_list(0)

    save_button = QPushButton("Save")
    layout.addWidget(save_button)

    def on_save_clicked():
        df_idx = combo.currentIndex()
        save_to_file(df_idx, app_instance, parent_widget)
        dialog.accept()

    save_button.clicked.connect(on_save_clicked)

    dialog.exec_()


def save_to_file(df_idx, app_instance, parent_widget=None):
    '''
    Save the selected data to a new text file.
    If fewer columns than the 8 available are loaded,
    columns of zeros will be added to maintain compatibility.

    Parameters
    ----------
    df_idx : int
        index of the selected dataframe
    app_instance : MainApp object
        instance of MainApp from main_window.py
    save_window : Toplevel object
        window to close after saving the data
    '''  

    dataframes   = app_instance.dataframes
    header_lines = app_instance.header_lines
   
    try:
        # Retrieve the selected DataFrame
        df = dataframes[df_idx]
       
        # Retrieve the header of the selected file
        header = header_lines[df_idx]
 
        # Dialog to choose the name of the new file
        file_path, _ = QFileDialog.getSaveFileName(
            parent_widget,
            "Save modified file",
            "",
            "Text file (*.txt);;CSV (*.csv);;Tutti i file (*)"
        )
    
        if not file_path:
            QMessageBox.warning(parent_widget, "Canceled", "Operation cancelled.")
            return

        save_header(app_instance, header, file_path)
        # Save the data in the new file in text format
        with open(file_path, "a", encoding='utf-8') as f:

            np.savetxt(f, np.array(df), delimiter="\t", fmt="%s")

        QMessageBox.information(parent_widget, "Success", f"Data successfully saved in:\n{file_path}")

    except Exception as e:
        QMessageBox.critical(parent_widget, "Error", f"Error while saving:\n{e}")

#==============================================================================================#
# Function that create a copy of a given file                                                  #
#==============================================================================================#

def duplicate_file(parent_widget=None):
    '''
    Duplicates a selected file by creating a copy named <original>_copy.ext.
    Opens a file dialog to select a file and creates a duplicate
    of it with '_copy' appended to the name.

    Parameters:
    -----------
    parent_widget : (QWidget, optional)
        The parent widget for dialogs. Defaults to None.
    '''
    file_path, _ = QFileDialog.getOpenFileName(
        parent_widget,
        "Select the file to duplicate",
        "",
        "Text Files (*.txt);;All Files (*)"
    )

    if not file_path:
        return

    base, ext = os.path.splitext(file_path)
    copy_path = f"{base}_copy{ext}"

    try:
        shutil.copy2(file_path, copy_path)
        QMessageBox.information(parent_widget, "Copy completed",
                                f"File copied as:\n{copy_path}")
    except Exception as e:
        QMessageBox.critical(parent_widget, "Error",
                             f"Error copying file:\n{e}")



# This file is part of HYLOA - HYsteresis LOop Analyzer.
# Copyright (C) 2024 Francesco Zeno Costanzo

# HYLOA is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# HYLOA is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with HYLOA. If not, see <https://www.gnu.org/licenses/>.

"""
Test log window widget.
"""
import os
import tempfile
import pytest
from PyQt5.QtCore import Qt
from PyQt5.QtGui import QTextCursor
from PyQt5.QtCore import QTimer


from hyloa.gui.log_window import LogWindow

class DummyApp:
    def __init__(self, logger_path=None):
        self.logger_path = logger_path


def test_initialization(qtbot):
    app = DummyApp()
    widget = LogWindow(app)
    qtbot.addWidget(widget)

    assert widget.isReadOnly()
    assert widget.styleSheet() != ""
    assert widget.timer.isActive()
    assert widget.last_line_count == 0


def test_update_log_no_path(qtbot):
    app = DummyApp(logger_path=None)
    widget = LogWindow(app)
    qtbot.addWidget(widget)

    widget.setPlainText("initial")
    widget.update_log()

    assert widget.toPlainText() == "initial"
    assert widget.last_line_count == 0


def test_update_log_new_lines(qtbot):
    # Create a tmp file
    with tempfile.NamedTemporaryFile(mode="w+", delete=False) as tmp:
        tmp.write("INFO - first line\n")
        tmp.write("DEBUG - second line\n")
        tmp_path = tmp.name

    app = DummyApp(logger_path=tmp_path)
    widget = LogWindow(app)
    qtbot.addWidget(widget)

    widget.update_log()

    # The displayed text should be cleaned after " - "
    assert "first line" in widget.toPlainText()
    assert "second line" in widget.toPlainText()
    assert widget.last_line_count == 2

    os.remove(tmp_path)


def test_update_log_no_new_lines(qtbot):
    with tempfile.NamedTemporaryFile(mode="w+", delete=False) as tmp:
        tmp.write("INFO - hello\n")
        tmp_path = tmp.name

    app = DummyApp(logger_path=tmp_path)
    widget = LogWindow(app)
    qtbot.addWidget(widget)

    widget.update_log()
    first_text = widget.toPlainText()

    # No new lines added
    widget.update_log()
    second_text = widget.toPlainText()

    assert first_text == second_text
    assert widget.last_line_count == 1

    os.remove(tmp_path)


def test_update_log_scroll_behavior(qtbot):
    with tempfile.NamedTemporaryFile(mode="w+", delete=False) as tmp:
        tmp.write("INFO - line1\n")
        tmp.write("INFO - line2\n")
        tmp.write("INFO - line3\n")
        tmp_path = tmp.name

    app = DummyApp(logger_path=tmp_path)
    widget = LogWindow(app)
    qtbot.addWidget(widget)

    # first update
    widget.update_log()

    # scroll to top
    sb = widget.verticalScrollBar()
    sb.setValue(0)

    with open(tmp_path, "a") as f:
        f.write("INFO - new line\n")

    widget.update_log()
    # scrollbar should remain at top
    assert sb.value() == 0

    os.remove(tmp_path)


def test_update_log_error_handling(qtbot):
    # Non-existent file path should trigger error handling
    app = DummyApp(logger_path="/nonexistent/path/log.txt")
    widget = LogWindow(app)
    qtbot.addWidget(widget)

    widget.update_log()

    assert "[Error]" in widget.toPlainText()
